from typing import TYPE_CHECKING

from slither.core.variables.variable import Variable
from slither.core.solidity_types.user_defined_type import UserDefinedType
from slither.core.solidity_types.mapping_type import MappingType
from slither.core.solidity_types.elementary_type import ElementaryType

from slither.core.declarations.structure import Structure

if TYPE_CHECKING:  # type: ignore
    from slither.core.declarations import Function


class LocalVariable(Variable):
    def __init__(self) -> None:
        super().__init__()
        self._location: str | None = None
        self._function: Function | None = None

    def set_function(self, function: "Function") -> None:
        self._function = function

    @property
    def function(self) -> "Function":
        assert self._function
        return self._function

    def set_location(self, loc: str) -> None:
        self._location = loc

    @property
    def location(self) -> str | None:
        """
            Variable Location
            Can be storage/memory or default
        Returns:
            (str)
        """
        return self._location

    @property
    def is_scalar(self) -> bool:
        return isinstance(self.type, ElementaryType) and not self.is_storage

    @property
    def is_storage(self) -> bool:
        """
            Return true if the variable is located in storage
            See https://solidity.readthedocs.io/en/v0.4.24/types.html?highlight=storage%20location#data-location
        Returns:
            (bool)
        """

        from slither.core.solidity_types.array_type import ArrayType

        if self.location == "memory":
            return False
        if self.location == "calldata":
            return False
        # Use by slithIR SSA
        if self.location == "reference_to_storage":
            return False
        if self.location == "storage":
            return True

        if isinstance(self.type, (ArrayType, MappingType)):
            return True

        if isinstance(self.type, UserDefinedType):
            return isinstance(self.type.type, Structure)

        return False

    @property
    def canonical_name(self) -> str:
        return f"{self.function.canonical_name}.{self.name}"



from typing import TYPE_CHECKING

from slither.core.source_mapping.source_mapping import SourceMapping
from slither.core.variables.local_variable import LocalVariable
from slither.utils.type import is_underlying_type_address

if TYPE_CHECKING:
    from slither.core.compilation_unit import SlitherCompilationUnit


class CustomError(SourceMapping):
    def __init__(self, compilation_unit: "SlitherCompilationUnit") -> None:
        super().__init__()
        self._name: str = ""
        self._parameters: list[LocalVariable] = []
        self._compilation_unit = compilation_unit

        self._solidity_signature: str | None = None
        self._full_name: str | None = None
        self._pattern = "error"

    @property
    def name(self) -> str:
        return self._name

    @name.setter
    def name(self, new_name: str) -> None:
        self._name = new_name

    @property
    def parameters(self) -> list[LocalVariable]:
        return self._parameters

    def add_parameters(self, p: "LocalVariable") -> None:
        self._parameters.append(p)

    @property
    def compilation_unit(self) -> "SlitherCompilationUnit":
        return self._compilation_unit

    # region Signature
    ###################################################################################
    ###################################################################################

    @staticmethod
    def _convert_type_for_solidity_signature(t: type | None) -> str:
        if is_underlying_type_address(t):
            return "address"
        return str(t)

    @property
    def solidity_signature(self) -> str:
        """
        Return a signature following the Solidity Standard
        Contract and converted into address
        :return: the solidity signature
        """
        # Ideally this should be an assert
        # But due to a logic limitation in the solc parsing (find_variable)
        # We need to raise an error if the custom error sig was not yet built
        # (set_solidity_sig was not called before find_variable)
        if self._solidity_signature is None:
            raise ValueError("Custom Error not yet built")
        return self._solidity_signature  # type: ignore

    def set_solidity_sig(self) -> None:
        """
        Function to be called once all the parameters have been set

        Returns:

        """
        parameters = [x.type for x in self.parameters if x.type]
        self._full_name = self.name + "(" + ",".join(map(str, parameters)) + ")"
        solidity_parameters = map(self._convert_type_for_solidity_signature, parameters)
        self._solidity_signature = self.name + "(" + ",".join(solidity_parameters) + ")"

    @property
    def full_name(self) -> str | None:
        """
        Return the error signature without
        converting contract into address
        :return: the error signature
        """
        if self._full_name is None:
            raise ValueError("Custom Error not yet built")
        return self._full_name

    # endregion
    ###################################################################################
    ###################################################################################

    def __str__(self) -> str:
        return "revert " + self.solidity_signature



import abc
from logging import Logger
from collections.abc import Callable

from slither.core.declarations import Contract
from slither.utils.colors import green, yellow, red
from slither.utils.comparable_enum import ComparableEnum
from slither.utils.output import Output, SupportedOutput


class IncorrectCheckInitialization(Exception):
    pass


class CheckClassification(ComparableEnum):
    HIGH = 0
    MEDIUM = 1
    LOW = 2
    INFORMATIONAL = 3
    UNIMPLEMENTED = 999


classification_colors: dict[CheckClassification, Callable[[str], str]] = {
    CheckClassification.INFORMATIONAL: green,
    CheckClassification.LOW: yellow,
    CheckClassification.MEDIUM: yellow,
    CheckClassification.HIGH: red,
}

classification_txt = {
    CheckClassification.INFORMATIONAL: "Informational",
    CheckClassification.LOW: "Low",
    CheckClassification.MEDIUM: "Medium",
    CheckClassification.HIGH: "High",
}

CHECK_INFO = list[str | SupportedOutput]


class AbstractCheck(metaclass=abc.ABCMeta):
    ARGUMENT = ""
    HELP = ""
    IMPACT: CheckClassification = CheckClassification.UNIMPLEMENTED

    WIKI = ""

    WIKI_TITLE = ""
    WIKI_DESCRIPTION = ""
    WIKI_EXPLOIT_SCENARIO = ""
    WIKI_RECOMMENDATION = ""

    REQUIRE_CONTRACT = False
    REQUIRE_PROXY = False
    REQUIRE_CONTRACT_V2 = False

    def __init__(
        self,
        logger: Logger,
        contract: Contract,
        proxy: Contract | None = None,
        contract_v2: Contract | None = None,
    ) -> None:
        self.logger = logger
        self.contract = contract
        self.proxy = proxy
        self.contract_v2 = contract_v2

        if not self.ARGUMENT:
            raise IncorrectCheckInitialization(f"NAME is not initialized {self.__class__.__name__}")

        if not self.HELP:
            raise IncorrectCheckInitialization(f"HELP is not initialized {self.__class__.__name__}")

        if not self.WIKI:
            raise IncorrectCheckInitialization(f"WIKI is not initialized {self.__class__.__name__}")

        if not self.WIKI_TITLE:
            raise IncorrectCheckInitialization(
                f"WIKI_TITLE is not initialized {self.__class__.__name__}"
            )

        if not self.WIKI_DESCRIPTION:
            raise IncorrectCheckInitialization(
                f"WIKI_DESCRIPTION is not initialized {self.__class__.__name__}"
            )

        if not self.WIKI_EXPLOIT_SCENARIO and self.IMPACT not in [
            CheckClassification.INFORMATIONAL
        ]:
            raise IncorrectCheckInitialization(
                f"WIKI_EXPLOIT_SCENARIO is not initialized {self.__class__.__name__}"
            )

        if not self.WIKI_RECOMMENDATION:
            raise IncorrectCheckInitialization(
                f"WIKI_RECOMMENDATION is not initialized {self.__class__.__name__}"
            )

        if self.REQUIRE_PROXY and self.REQUIRE_CONTRACT_V2:
            # This is not a fundatemenal issues
            # But it requires to change __main__ to avoid running two times the detectors
            txt = f"REQUIRE_PROXY and REQUIRE_CONTRACT_V2 needs change in __main___ {self.__class__.__name__}"
            raise IncorrectCheckInitialization(txt)

        if self.IMPACT not in [
            CheckClassification.LOW,
            CheckClassification.MEDIUM,
            CheckClassification.HIGH,
            CheckClassification.INFORMATIONAL,
        ]:
            raise IncorrectCheckInitialization(
                f"IMPACT is not initialized {self.__class__.__name__}"
            )

        if self.REQUIRE_CONTRACT_V2 and contract_v2 is None:
            raise IncorrectCheckInitialization(
                f"ContractV2 is not initialized {self.__class__.__name__}"
            )

        if self.REQUIRE_PROXY and proxy is None:
            raise IncorrectCheckInitialization(
                f"Proxy is not initialized {self.__class__.__name__}"
            )

    @abc.abstractmethod
    def _check(self) -> list[Output]:
        """TODO Documentation"""
        return []

    def check(self) -> list[dict]:
        all_outputs = self._check()
        # Keep only dictionaries
        all_results = [r.data for r in all_outputs]
        if all_results:
            if self.logger:
                info = "\n"
                for result in all_results:
                    info += result["description"]
                info += f"Reference: {self.WIKI}"
                self._log(info)
        return all_results

    def generate_result(
        self,
        info: CHECK_INFO,
        additional_fields: dict | None = None,
    ) -> Output:
        output = Output(
            info, additional_fields, markdown_root=self.contract.compilation_unit.core.markdown_root
        )

        output.data["check"] = self.ARGUMENT

        return output

    def _log(self, info: str) -> None:
        if self.logger:
            self.logger.info(self.color(info))

    @property
    def color(self) -> Callable[[str], str]:
        return classification_colors[self.IMPACT]



from slither.slithir.operations import Binary, BinaryType
from slither.tools.mutator.utils.patch import create_patch_with_line
from slither.tools.mutator.mutators.abstract_mutator import AbstractMutator

relational_operators = [
    BinaryType.LESS,
    BinaryType.GREATER,
    BinaryType.LESS_EQUAL,
    BinaryType.GREATER_EQUAL,
    BinaryType.EQUAL,
    BinaryType.NOT_EQUAL,
]


class ROR(AbstractMutator):
    NAME = "ROR"
    HELP = "Relational Operator Replacement"

    def _mutate(self) -> dict:
        result: dict = {}

        for function in self.contract.functions_and_modifiers_declared:
            if not self.should_mutate_function(function):
                continue
            for node in function.nodes:
                if not self.should_mutate_node(node):
                    continue
                for ir in node.irs:
                    if isinstance(ir, Binary) and ir.type in relational_operators:
                        if (
                            str(ir.variable_left.type) != "address"
                            and str(ir.variable_right) != "address"
                        ):
                            alternative_ops = relational_operators[:]
                            alternative_ops.remove(ir.type)
                            for op in alternative_ops:
                                # Get the string
                                start = ir.expression.source_mapping.start
                                stop = start + ir.expression.source_mapping.length
                                old_str = node.source_mapping.content
                                line_no = node.source_mapping.lines
                                # Replace the expression with true
                                new_str = f"{old_str.split(ir.type.value)[0]}{op.value}{old_str.split(ir.type.value)[1]}"
                                create_patch_with_line(
                                    result,
                                    self.in_file,
                                    start,
                                    stop,
                                    old_str,
                                    new_str,
                                    line_no[0],
                                )
        return result



from slither.core.declarations.contract_level import ContractLevel
from slither.core.declarations import Structure


class StructureContract(Structure, ContractLevel):
    def is_declared_by(self, contract):
        """
        Check if the element is declared by the contract
        :param contract:
        :return:
        """
        return self.contract == contract



from slither.core.declarations import Function, FunctionContract, Contract
from slither.core.declarations.structure import Structure
from slither.core.solidity_types.array_type import ArrayType
from slither.core.solidity_types.user_defined_type import UserDefinedType
from slither.core.variables.variable import Variable
from slither.detectors.abstract_detector import (
    AbstractDetector,
    DetectorClassification,
    ALL_SOLC_VERSIONS_04,
    ALL_SOLC_VERSIONS_05,
    make_solc_versions,
)
from slither.formatters.functions.external_function import custom_format
from slither.slithir.operations import InternalDynamicCall
from slither.utils.output import Output


class ExternalFunction(AbstractDetector):
    """
    Detect public function that could be declared as external

    IMPROVEMENT: Add InternalDynamicCall check
    https://github.com/trailofbits/slither/pull/53#issuecomment-432809950
    """

    ARGUMENT = "external-function"
    HELP = "Public function that could be declared external"
    IMPACT = DetectorClassification.OPTIMIZATION
    CONFIDENCE = DetectorClassification.HIGH

    WIKI = "https://github.com/crytic/slither/wiki/Detector-Documentation#public-function-that-could-be-declared-external"

    WIKI_TITLE = "Public function that could be declared external"
    WIKI_DESCRIPTION = "`public` functions that are never called by the contract should be declared `external`, and its immutable parameters should be located in `calldata` to save gas."
    WIKI_RECOMMENDATION = "Use the `external` attribute for functions never called from the contract, and change the location of immutable parameters to `calldata` to save gas."

    VULNERABLE_SOLC_VERSIONS = (
        ALL_SOLC_VERSIONS_04 + ALL_SOLC_VERSIONS_05 + make_solc_versions(6, 0, 8)
    )

    @staticmethod
    def detect_functions_called(contract: Contract) -> list[Function]:
        """Returns a list of InternallCall, SolidityCall
            calls made in a function

        Returns:
            (list): List of all InternallCall, SolidityCall
        """
        result = []

        # Obtain all functions reachable by this contract.
        for func in contract.all_functions_called:
            if not isinstance(func, Function):
                continue

            # Loop through all internal and solidity calls in the function, add them to a list.
            for ir in func.internal_calls + func.solidity_calls:
                result.append(ir.function)

        return result

    @staticmethod
    def _contains_internal_dynamic_call(contract: Contract) -> bool:
        """
        Checks if a contract contains a dynamic call either in a direct definition, or through inheritance.

        Returns:
            (boolean): True if this contract contains a dynamic call (including through inheritance).
        """
        for func in contract.all_functions_called:
            if not isinstance(func, Function):
                continue
            for node in func.nodes:
                for ir in node.irs:
                    if isinstance(ir, (InternalDynamicCall)):
                        return True
        return False

    @staticmethod
    def get_base_most_function(function: FunctionContract) -> FunctionContract:
        """
        Obtains the base function definition for the provided function. This could be used to obtain the original
        definition of a function, if the provided function is an override.

        Returns:
            (function): Returns the base-most function of a provided function. (The original definition).
        """
        # Loop through the list of inherited contracts and this contract, to find the first function instance which
        # matches this function's signature. Note here that `inheritance` is in order from most basic to most extended.
        for contract in function.contract.inheritance + [function.contract]:
            # Loop through the functions not inherited (explicitly defined in this contract).
            for f in contract.functions_declared:
                # If it matches names, this is the base most function.
                if f.full_name == function.full_name:
                    return f

        # Somehow we couldn't resolve it, which shouldn't happen, as the provided function should be found if we could
        # not find some any more basic.

        raise Exception("Could not resolve the base-most function for the provided function.")

    @staticmethod
    def get_all_function_definitions(
        base_most_function: FunctionContract,
    ) -> list[FunctionContract]:
        """
        Obtains all function definitions given a base-most function. This includes the provided function, plus any
        overrides of that function.

        Returns:
            (list): Returns any the provided function and any overriding functions defined for it.
        """
        # We assume the provided function is the base-most function, so we check all derived contracts
        # for a redefinition
        return [base_most_function] + [
            function
            for derived_contract in base_most_function.contract.derived_contracts
            for function in derived_contract.functions
            if function.full_name == base_most_function.full_name
            and isinstance(function, FunctionContract)
        ]

    @staticmethod
    def function_parameters_written(function: Function) -> bool:
        return any(p in function.variables_written for p in function.parameters)

    @staticmethod
    def is_reference_type(parameter: Variable) -> bool:
        parameter_type = parameter.type
        if isinstance(parameter_type, ArrayType):
            return True
        if isinstance(parameter_type, UserDefinedType) and isinstance(
            parameter_type.type, Structure
        ):
            return True
        if str(parameter_type) in ["bytes", "string"]:
            return True
        return False

    def _detect(self) -> list[Output]:
        results: list[Output] = []

        # Create a set to track contracts with dynamic calls. All contracts with dynamic calls could potentially be
        # calling functions internally, and thus we can't assume any function in such contracts isn't called by them.
        dynamic_call_contracts: set[Contract] = set()

        # Create a completed functions set to skip over functions already processed (any functions which are the base
        # of, or override hierarchically are processed together).
        completed_functions: set[Function] = set()

        # First we build our set of all contracts with dynamic calls
        for contract in self.contracts:
            if self._contains_internal_dynamic_call(contract):
                dynamic_call_contracts.add(contract)

        # Loop through all contracts
        for contract in self.contracts:
            # Filter false-positives: Immediately filter this contract if it's in blacklist
            if contract in dynamic_call_contracts:
                continue

            # Next we'll want to loop through all functions defined directly in this contract.
            for function in contract.functions_declared:
                # If all of the function arguments are non-reference type or calldata, we skip it.
                reference_args = []
                for arg in function.parameters:
                    if self.is_reference_type(arg) and arg.location == "memory":
                        reference_args.append(arg)
                if len(reference_args) == 0:
                    continue

                # If the function is a constructor, or is public, we skip it.
                if function.is_constructor or function.visibility != "public":
                    continue

                # Optimization: If this function has already been processed, we stop.
                if function in completed_functions:
                    continue

                # If the function has parameters which are written-to in function body, we skip
                # because parameters of external functions will be allocated in calldata region which is immutable
                if self.function_parameters_written(function):
                    continue

                # Get the base-most function to know our origin of this function.
                base_most_function = self.get_base_most_function(function)

                # Get all possible contracts which can call this function (or an override).
                all_possible_sources = [
                    base_most_function.contract
                ] + base_most_function.contract.derived_contracts

                # Get all function signatures (overloaded and not), mark as completed and we process them now.
                # Note: We mark all function definitions as the same, as they must all share visibility to override.
                all_function_definitions = set(
                    self.get_all_function_definitions(base_most_function)
                )
                completed_functions = completed_functions.union(all_function_definitions)

                # Filter false-positives: Determine if any of these sources have dynamic calls, if so, flag all of these
                # function definitions, and then flag all functions in all contracts that make dynamic calls.
                sources_with_dynamic_calls = set(all_possible_sources) & dynamic_call_contracts
                if sources_with_dynamic_calls:
                    functions_in_dynamic_call_sources = {
                        f
                        for dyn_contract in sources_with_dynamic_calls
                        for f in dyn_contract.functions
                        if not f.is_constructor
                    }
                    completed_functions = completed_functions.union(
                        functions_in_dynamic_call_sources
                    )
                    continue

                # Detect all functions called in each source, if any match our current signature, we skip
                # otherwise, this is a candidate (in all sources) to be changed visibility for.
                is_called = False
                for possible_source in all_possible_sources:
                    functions_called = self.detect_functions_called(possible_source)
                    if set(functions_called) & all_function_definitions:
                        is_called = True
                        break

                # If any of this function's definitions are called, we skip it.
                if is_called:
                    continue

                # As we collect all shadowed functions in get_all_function_definitions
                # Some function coming from a base might already been declared as external
                all_function_definitions: list[FunctionContract] = [
                    f
                    for f in all_function_definitions
                    if isinstance(f, FunctionContract)
                    and f.visibility == "public"
                    and f.contract == f.contract_declarer
                ]
                if all_function_definitions:
                    all_function_definitions = sorted(
                        all_function_definitions, key=lambda x: x.canonical_name
                    )
                    function_definition = all_function_definitions[0]
                    all_function_definitions = all_function_definitions[1:]

                    info = [f"{function_definition.full_name} should be declared external:\n"]
                    info += ["\t- ", function_definition, "\n"]
                    if self.compilation_unit.solc_version >= "0.5.":
                        info += [
                            "Moreover, the following function parameters should change its data location:\n"
                        ]
                        for reference_arg in reference_args:
                            info += [f"{reference_arg} location should be calldata\n"]
                    for other_function_definition in all_function_definitions:
                        info += ["\t- ", other_function_definition, "\n"]

                    res = self.generate_result(info)

                    results.append(res)

        return results

    @staticmethod
    def _format(slither, result):
        custom_format(slither, result)



from typing import TYPE_CHECKING
from collections import deque
from slither.core.declarations.solidity_variables import (
    SOLIDITY_VARIABLES_COMPOSED,
    SolidityVariableComposed,
)
from slither.core.declarations import SolidityFunction, FunctionContract
from slither.core.variables.state_variable import StateVariable
from slither.core.expressions import (
    CallExpression,
    ElementaryTypeNameExpression,
    Identifier,
    IndexAccess,
    Literal,
    MemberAccess,
    SelfIdentifier,
    TupleExpression,
    TypeConversion,
    UnaryOperation,
    UnaryOperationType,
)
from slither.core.expressions.assignment_operation import (
    AssignmentOperation,
    AssignmentOperationType,
)
from slither.core.expressions.binary_operation import (
    BinaryOperation,
    BinaryOperationType,
)
from slither.core.solidity_types import (
    ArrayType,
    ElementaryType,
    UserDefinedType,
)
from slither.core.declarations.contract import Contract
from slither.vyper_parsing.expressions.find_variable import find_variable
from slither.vyper_parsing.type_parsing import parse_type
from slither.all_exceptions import ParsingError
from slither.vyper_parsing.ast.types import (
    Int,
    Call,
    Attribute,
    Name,
    Tuple,
    Hex,
    BinOp,
    Str,
    Assert,
    Compare,
    UnaryOp,
    Subscript,
    NameConstant,
    VyDict,
    Bytes,
    BoolOp,
    Assign,
    AugAssign,
    VyList,
    Raise,
    ASTNode,
)

if TYPE_CHECKING:
    from slither.core.expressions.expression import Expression


def vars_to_typestr(rets: list["Expression"] | None) -> str:
    if rets is None:
        return "tuple()"
    if len(rets) == 1:
        return str(rets[0].type)
    return f"tuple({','.join(str(ret.type) for ret in rets)})"


def parse_expression(
    expression: ASTNode, caller_context: FunctionContract | Contract
) -> "Expression":
    if isinstance(expression, Int):
        literal = Literal(str(expression.value), ElementaryType("uint256"))
        literal.set_offset(expression.src, caller_context.compilation_unit)
        return literal

    if isinstance(expression, Hex):
        # TODO this is an implicit conversion and could potentially be bytes20 or other? https://github.com/vyperlang/vyper/issues/3580
        literal = Literal(str(expression.value), ElementaryType("address"))
        literal.set_offset(expression.src, caller_context.compilation_unit)
        return literal

    if isinstance(expression, Str):
        literal = Literal(str(expression.value), ElementaryType("string"))
        literal.set_offset(expression.src, caller_context.compilation_unit)
        return literal

    if isinstance(expression, Bytes):
        literal = Literal(str(expression.value), ElementaryType("bytes"))
        literal.set_offset(expression.src, caller_context.compilation_unit)
        return literal

    if isinstance(expression, NameConstant):
        assert str(expression.value) in ["True", "False"]
        literal = Literal(str(expression.value), ElementaryType("bool"))
        literal.set_offset(expression.src, caller_context.compilation_unit)
        return literal

    if isinstance(expression, Call):
        called = parse_expression(expression.func, caller_context)
        if isinstance(called, Identifier) and isinstance(called.value, SolidityFunction):
            if called.value.name == "empty()":
                type_to = parse_type(expression.args[0], caller_context)
                # TODO figure out how to represent this type argument
                parsed_expr = CallExpression(called, [], str(type_to))
                parsed_expr.set_offset(expression.src, caller_context.compilation_unit)
                return parsed_expr

            if called.value.name == "convert()":
                arg = parse_expression(expression.args[0], caller_context)
                type_to = parse_type(expression.args[1], caller_context)
                parsed_expr = TypeConversion(arg, type_to)
                parsed_expr.set_offset(expression.src, caller_context.compilation_unit)
                return parsed_expr

            if called.value.name == "min_value()":
                type_to = parse_type(expression.args[0], caller_context)
                member_type = str(type_to)
                # TODO return Literal
                parsed_expr = MemberAccess(
                    "min",
                    member_type,
                    CallExpression(
                        Identifier(SolidityFunction("type()")),
                        [ElementaryTypeNameExpression(type_to)],
                        member_type,
                    ),
                )
                parsed_expr.set_offset(expression.src, caller_context.compilation_unit)
                return parsed_expr

            if called.value.name == "max_value()":
                type_to = parse_type(expression.args[0], caller_context)
                member_type = str(type_to)
                # TODO return Literal
                parsed_expr = MemberAccess(
                    "max",
                    member_type,
                    CallExpression(
                        Identifier(SolidityFunction("type()")),
                        [ElementaryTypeNameExpression(type_to)],
                        member_type,
                    ),
                )
                parsed_expr.set_offset(expression.src, caller_context.compilation_unit)
                return parsed_expr

            # `raw_call` and `send` are treated specially in order to force `extract_tmp_call` to treat this as a `HighLevelCall` which will be converted
            # to a `LowLevelCall` by `convert_to_low_level`. This is an artifact of the late conversion of Solidity...
            if called.value.name == "raw_call()":
                args = [parse_expression(a, caller_context) for a in expression.args]
                call = CallExpression(
                    MemberAccess("raw_call", "tuple(bool,bytes32)", args[0]),
                    args[1:],
                    "tuple(bool,bytes32)",
                )
                call.set_offset(expression.src, caller_context.compilation_unit)
                call.call_value = next(
                    iter(
                        parse_expression(x.value, caller_context)
                        for x in expression.keywords
                        if x.arg == "value"
                    ),
                    None,
                )
                call.call_gas = next(
                    iter(
                        parse_expression(x.value, caller_context)
                        for x in expression.keywords
                        if x.arg == "gas"
                    ),
                    None,
                )
                # TODO handle `max_outsize` keyword

                return call

            if called.value.name == "send()":
                args = [parse_expression(a, caller_context) for a in expression.args]
                call = CallExpression(
                    MemberAccess("send", "tuple()", args[0]),
                    args[1:],
                    "tuple()",
                )
                call.set_offset(expression.src, caller_context.compilation_unit)

                return call

        if expression.args and isinstance(expression.args[0], VyDict):
            arguments = []
            for val in expression.args[0].values:
                arguments.append(parse_expression(val, caller_context))
        else:
            arguments = [parse_expression(a, caller_context) for a in expression.args]

        rets = None
        # Since the AST lacks the type of the return values, we recover it. https://github.com/vyperlang/vyper/issues/3581
        if isinstance(called, Identifier):
            if isinstance(called.value, FunctionContract):
                rets = called.value.returns
                # Default arguments are not represented in the AST, so we recover them as well.

                if called.value._default_args_as_expressions and len(arguments) < len(
                    called.value.parameters
                ):
                    arguments.extend(
                        [
                            parse_expression(x, caller_context)
                            for x in called.value._default_args_as_expressions
                        ]
                    )

            elif isinstance(called.value, SolidityFunction):
                rets = called.value.return_type

            elif isinstance(called.value, Contract):
                # Type conversions are not explicitly represented in the AST e.g. converting address to contract/ interface,
                # so we infer that a type conversion is occurring if `called` is a `Contract` type. https://github.com/vyperlang/vyper/issues/3580
                type_to = parse_type(expression.func, caller_context)
                parsed_expr = TypeConversion(arguments[0], type_to)
                parsed_expr.set_offset(expression.src, caller_context.compilation_unit)
                return parsed_expr

        elif isinstance(called, MemberAccess) and called.type is not None:
            # (recover_type_2) Propagate the type collected to the `CallExpression`
            # see recover_type_1
            rets = [called]

        type_str = vars_to_typestr(rets)
        parsed_expr = CallExpression(called, arguments, type_str)
        parsed_expr.set_offset(expression.src, caller_context.compilation_unit)
        return parsed_expr

    if isinstance(expression, Attribute):
        member_name = expression.attr
        if isinstance(expression.value, Name):
            # TODO this is ambiguous because it could be a state variable or a call to balance https://github.com/vyperlang/vyper/issues/3582
            if expression.value.id == "self" and member_name != "balance":
                var = find_variable(member_name, caller_context, is_self=True)
                parsed_expr = SelfIdentifier(var)
                parsed_expr.set_offset(expression.src, caller_context.compilation_unit)
                var.references.append(parsed_expr.source_mapping)
                return parsed_expr

            expr = parse_expression(expression.value, caller_context)
            # TODO this is ambiguous because it could be a type conversion of an interface or a member access
            # see https://github.com/vyperlang/vyper/issues/3580 and ttps://github.com/vyperlang/vyper/issues/3582
            if expression.attr == "address":
                parsed_expr = TypeConversion(expr, ElementaryType("address"))
                parsed_expr.set_offset(expression.src, caller_context.compilation_unit)
                return parsed_expr

            member_access = MemberAccess(member_name, None, expr)

            if str(member_access) in SOLIDITY_VARIABLES_COMPOSED:
                parsed_expr = Identifier(SolidityVariableComposed(str(member_access)))
                parsed_expr.set_offset(expression.src, caller_context.compilation_unit)
                return parsed_expr

        else:
            expr = parse_expression(expression.value, caller_context)
            member_name_ret_type = None
            # (recover_type_1) This may be a call to an interface and we don't have the return types,
            # so we see if there's a function identifier with `member_name` and propagate the type to
            # its enclosing `CallExpression`. https://github.com/vyperlang/vyper/issues/3581
            if (
                isinstance(expr, Identifier)
                and isinstance(expr.value, StateVariable)
                and isinstance(expr.value.type, UserDefinedType)
                and isinstance(expr.value.type.type, Contract)
            ):
                # If we access a member of an interface, needs to be interface instead of self namespace
                var = find_variable(member_name, expr.value.type.type)
                if isinstance(var, FunctionContract):
                    rets = var.returns
                    member_name_ret_type = vars_to_typestr(rets)

            if (
                isinstance(expr, TypeConversion)
                and isinstance(expr.type, UserDefinedType)
                and isinstance(expr.type.type, Contract)
            ):
                # If we access a member of an interface, needs to be interface instead of self namespace
                var = find_variable(member_name, expr.type.type)
                if isinstance(var, FunctionContract):
                    rets = var.returns
                    member_name_ret_type = vars_to_typestr(rets)

            member_access = MemberAccess(member_name, member_name_ret_type, expr)

        member_access.set_offset(expression.src, caller_context.compilation_unit)
        return member_access

    if isinstance(expression, Name):
        var = find_variable(expression.id, caller_context)
        parsed_expr = Identifier(var)
        parsed_expr.set_offset(expression.src, caller_context.compilation_unit)
        return parsed_expr

    if isinstance(expression, Assign):
        lhs = parse_expression(expression.target, caller_context)
        rhs = parse_expression(expression.value, caller_context)
        parsed_expr = AssignmentOperation(lhs, rhs, AssignmentOperationType.ASSIGN, None)
        parsed_expr.set_offset(expression.src, caller_context.compilation_unit)
        return parsed_expr

    if isinstance(expression, AugAssign):
        lhs = parse_expression(expression.target, caller_context)
        rhs = parse_expression(expression.value, caller_context)

        op = AssignmentOperationType.get_type(expression.op)
        parsed_expr = AssignmentOperation(lhs, rhs, op, None)
        parsed_expr.set_offset(expression.src, caller_context.compilation_unit)
        return parsed_expr

    if isinstance(expression, (Tuple, VyList)):
        tuple_vars = [parse_expression(x, caller_context) for x in expression.elements]
        parsed_expr = TupleExpression(tuple_vars)
        parsed_expr.set_offset(expression.src, caller_context.compilation_unit)
        return parsed_expr

    if isinstance(expression, UnaryOp):
        operand = parse_expression(expression.operand, caller_context)
        op = UnaryOperationType.get_type(
            expression.op, isprefix=True
        )  # TODO does vyper have postfix?

        parsed_expr = UnaryOperation(operand, op)
        parsed_expr.set_offset(expression.src, caller_context.compilation_unit)
        return parsed_expr

    if isinstance(expression, Compare):
        lhs = parse_expression(expression.left, caller_context)

        # We assume left operand in membership comparison cannot be Array type
        if expression.op in ["In", "NotIn"]:
            # If we see a membership operator e.g. x in [foo(), bar()], we convert it to logical operations
            # like (x == foo() || x == bar()) or (x != foo() && x != bar()) for "not in"
            # TODO consider rewriting as if-else to accurately represent the precedence of potential side-effects

            conditions = deque()
            rhs = parse_expression(expression.right, caller_context)
            is_tuple = isinstance(rhs, TupleExpression)
            is_array = isinstance(rhs, Identifier) and isinstance(rhs.value.type, ArrayType)
            if is_array:
                assert rhs.value.type.is_fixed_array, (
                    "Dynamic arrays are not supported in comparison operators"
                )
            if is_tuple or is_array:
                length = len(rhs.expressions) if is_tuple else rhs.value.type.length_value.value
                inner_op = (
                    BinaryOperationType.get_type("!=")
                    if expression.op == "NotIn"
                    else BinaryOperationType.get_type("==")
                )
                for i in range(length):
                    elem_expr = (
                        rhs.expressions[i]
                        if is_tuple
                        else IndexAccess(rhs, Literal(str(i), ElementaryType("uint256")))
                    )
                    elem_expr.set_offset(rhs.source_mapping, caller_context.compilation_unit)
                    parsed_expr = BinaryOperation(lhs, elem_expr, inner_op)
                    parsed_expr.set_offset(lhs.source_mapping, caller_context.compilation_unit)
                    conditions.append(parsed_expr)

                outer_op = (
                    BinaryOperationType.get_type("&&")
                    if expression.op == "NotIn"
                    else BinaryOperationType.get_type("||")
                )
                while len(conditions) > 1:
                    lhs = conditions.pop()
                    rhs = conditions.pop()

                    conditions.appendleft(BinaryOperation(lhs, rhs, outer_op))

                return conditions.pop()

            # enum type membership check https://docs.vyperlang.org/en/stable/types.html?h#id18
            is_member_op = (
                BinaryOperationType.get_type("==")
                if expression.op == "NotIn"
                else BinaryOperationType.get_type("!=")
            )
            # If all bits are cleared, then the lhs is not a member of the enum
            # This allows representing membership in multiple enum members
            # For example, if enum Foo has members A (1), B (2), and C (4), then
            # (x in [Foo.A, Foo.B]) is equivalent to (x & (Foo.A | Foo.B) != 0),
            # where (Foo.A | Foo.B) evaluates to 3.
            # Thus, when x is 3, (x & (Foo.A | Foo.B) != 0) is true.
            enum_bit_mask = BinaryOperation(
                TypeConversion(lhs, ElementaryType("uint256")),
                TypeConversion(rhs, ElementaryType("uint256")),
                BinaryOperationType.get_type("&"),
            )
            membership_check = BinaryOperation(
                enum_bit_mask, Literal("0", ElementaryType("uint256")), is_member_op
            )
            membership_check.set_offset(lhs.source_mapping, caller_context.compilation_unit)
            return membership_check

        # a regular logical operator
        rhs = parse_expression(expression.right, caller_context)
        op = BinaryOperationType.get_type(expression.op)

        parsed_expr = BinaryOperation(lhs, rhs, op)
        parsed_expr.set_offset(expression.src, caller_context.compilation_unit)
        return parsed_expr

    if isinstance(expression, BinOp):
        lhs = parse_expression(expression.left, caller_context)
        rhs = parse_expression(expression.right, caller_context)

        op = BinaryOperationType.get_type(expression.op)
        parsed_expr = BinaryOperation(lhs, rhs, op)
        parsed_expr.set_offset(expression.src, caller_context.compilation_unit)
        return parsed_expr

    if isinstance(expression, Assert):
        # Treat assert the same as a Solidity `require`.
        # TODO rename from `SolidityFunction` to `Builtin`?
        type_str = "tuple()"
        if expression.msg is None:
            func = SolidityFunction("require(bool)")
            args = [parse_expression(expression.test, caller_context)]
        else:
            func = SolidityFunction("require(bool,string)")
            args = [
                parse_expression(expression.test, caller_context),
                parse_expression(expression.msg, caller_context),
            ]

        parsed_expr = CallExpression(Identifier(func), args, type_str)
        parsed_expr.set_offset(expression.src, caller_context.compilation_unit)
        return parsed_expr

    if isinstance(expression, Subscript):
        left_expression = parse_expression(expression.value, caller_context)
        right_expression = parse_expression(expression.slice.value, caller_context)
        parsed_expr = IndexAccess(left_expression, right_expression)
        parsed_expr.set_offset(expression.src, caller_context.compilation_unit)
        return parsed_expr

    if isinstance(expression, BoolOp):
        lhs = parse_expression(expression.values[0], caller_context)
        rhs = parse_expression(expression.values[1], caller_context)

        op = BinaryOperationType.get_type(expression.op)
        parsed_expr = BinaryOperation(lhs, rhs, op)
        parsed_expr.set_offset(expression.src, caller_context.compilation_unit)
        return parsed_expr

    if isinstance(expression, Raise):
        type_str = "tuple()"
        func = (
            SolidityFunction("revert()")
            if expression.exc is None
            else SolidityFunction("revert(string)")
        )
        args = [] if expression.exc is None else [parse_expression(expression.exc, caller_context)]

        parsed_expr = CallExpression(Identifier(func), args, type_str)
        parsed_expr.set_offset(expression.src, caller_context.compilation_unit)
        return parsed_expr

    raise ParsingError(f"Expression not parsed {expression}")



import os
import shutil
import tempfile
from pathlib import Path
from contextlib import contextmanager
from filelock import FileLock
from solc_select import solc_select
import pytest
from slither import Slither


def pytest_configure(config):
    """Create a temporary directory for the tests to use."""
    if is_master():
        config.stash["shared_directory"] = tempfile.mkdtemp()


def pytest_unconfigure(config):
    """Remove the temporary directory after the tests are done."""
    if is_master():
        shutil.rmtree(config.stash["shared_directory"])


def pytest_configure_node(node):
    """Configure each worker node with the shared directory."""
    node.workerinput["shared_directory"] = node.config.stash["shared_directory"]


def is_master():
    """Returns True if the current process is the master process (which does not have a worker id)."""
    return os.environ.get("PYTEST_XDIST_WORKER") is None


@pytest.fixture
def shared_directory(request):
    """Returns the shared directory for the current process."""
    if is_master():
        return request.config.stash["shared_directory"]
    return request.config.workerinput["shared_directory"]


@pytest.fixture
def solc_binary_path(shared_directory):
    """
    Returns the path to the solc binary for the given version.
    If the binary is not installed, it will be installed.
    """

    def inner(version):
        lock = FileLock(f"{shared_directory}/{version}.lock", timeout=60)
        with lock:
            if not solc_select.artifact_path(version).exists():
                print("Installing solc version", version)
                solc_select.install_artifacts([version])
        return solc_select.artifact_path(version).as_posix()

    return inner


@pytest.fixture
def slither_from_solidity_source(solc_binary_path):
    @contextmanager
    def inner(source_code: str, solc_version: str = "0.8.19", legacy: bool = False):
        """Yields a Slither instance using source_code string and solc_version.
        Creates a temporary file and compiles with solc_version.
        """

        fname = ""
        try:
            with tempfile.NamedTemporaryFile(mode="w", suffix=".sol", delete=False) as f:
                fname = f.name
                f.write(source_code)
            solc_path = solc_binary_path(solc_version)
            yield Slither(fname, solc=solc_path, force_legacy=legacy)
        finally:
            Path(fname).unlink()

    return inner


@pytest.fixture
def slither_from_vyper_source():
    @contextmanager
    def inner(source_code: str):
        """Yields a Slither instance using source_code string.
        Creates a temporary file and compiles with vyper.
        """

        fname = ""
        try:
            with tempfile.NamedTemporaryFile(mode="w", suffix=".vy", delete=False) as f:
                fname = f.name
                f.write(source_code)
            yield Slither(fname)
        finally:
            Path(fname).unlink()

    return inner



import logging

from slither.core.declarations import Contract
from slither.utils import output

logger = logging.getLogger("Slither-conformance")


def approval_race_condition(contract: Contract, ret: dict[str, list]) -> None:
    increaseAllowance = contract.get_function_from_signature("increaseAllowance(address,uint256)")

    if not increaseAllowance:
        increaseAllowance = contract.get_function_from_signature(
            "safeIncreaseAllowance(address,uint256)"
        )

    if increaseAllowance:
        txt = f"\t[✓] {contract.name} has {increaseAllowance.full_name}"
        logger.info(txt)
    else:
        txt = f"\t[ ] {contract.name} is not protected for the ERC20 approval race condition"
        logger.info(txt)

        lack_of_erc20_race_condition_protection = output.Output(txt)
        lack_of_erc20_race_condition_protection.add(contract)
        ret["lack_of_erc20_race_condition_protection"].append(
            lack_of_erc20_race_condition_protection.data
        )


def check_erc20(
    contract: Contract, ret: dict[str, list], explored: bool | None = None
) -> dict[str, list]:
    if explored is None:
        explored = set()

    explored.add(contract)

    approval_race_condition(contract, ret)

    for derived_contract in contract.derived_contracts:
        check_erc20(derived_contract, ret, explored)

    return ret



from slither.detectors.abstract_detector import AbstractDetector, DetectorClassification


class Example(AbstractDetector):
    """
    Documentation
    """

    ARGUMENT = "mydetector"  # slither will launch the detector with slither.py --mydetector
    HELP = "Help printed by slither"
    IMPACT = DetectorClassification.HIGH
    CONFIDENCE = DetectorClassification.HIGH

    WIKI = "https://www.example.com/#example-detector"

    WIKI_TITLE = "example detector"
    WIKI_DESCRIPTION = "This is an example detector that always generates a finding"
    WIKI_EXPLOIT_SCENARIO = "Scenario goes here"
    WIKI_RECOMMENDATION = "Customize the detector"

    def _detect(self):
        info = "This is an example!"

        json = self.generate_result(info)

        return [json]



#!/usr/bin/env python3

import argparse
import cProfile
import glob
import inspect
import json
import logging
import os
import pstats
import sys
import traceback
from importlib import metadata
from typing import Any
from collections.abc import Sequence

from crytic_compile import cryticparser, CryticCompile, compile_all, is_supported
from crytic_compile.platform.etherscan import SUPPORTED_NETWORK
from crytic_compile.platform.standard import generate_standard_export

from slither.detectors import all_detectors
from slither.detectors.abstract_detector import AbstractDetector, DetectorClassification
from slither.exceptions import SlitherException
from slither.printers import all_printers
from slither.printers.abstract_printer import AbstractPrinter
from slither.slither import Slither
from slither.utils.colors import red, set_colorization_enabled
from slither.utils.command_line import (
    DEFAULT_JSON_OUTPUT_TYPES,
    JSON_OUTPUT_TYPES,
    FailOnLevel,
    check_and_sanitize_markdown_root,
    defaults_flag_in_config,
    output_detectors,
    output_detectors_json,
    output_printers,
    output_printers_json,
    output_results_to_markdown,
    output_to_markdown,
    output_wiki,
    read_config_file,
)
from slither.utils.output import (
    ZIP_TYPES_ACCEPTED,
    Output,
    output_to_json,
    output_to_sarif,
    output_to_zip,
)
from slither.utils.output_capture import StandardOutputCapture


logging.basicConfig()
logger = logging.getLogger("Slither")


###################################################################################
###################################################################################
# region Process functions
###################################################################################
###################################################################################


def process_single(
    target: str | CryticCompile,
    args: argparse.Namespace,
    detector_classes: list[type[AbstractDetector]],
    printer_classes: list[type[AbstractPrinter]],
) -> tuple[Slither, list[dict], list[Output], int]:
    """
    The core high-level code for running Slither static analysis.

    Returns:
        list(result), int: Result list and number of contracts analyzed
    """
    ast = "--ast-compact-json"
    if args.legacy_ast:
        ast = "--ast-json"
    slither = Slither(target, ast_format=ast, **vars(args))

    if args.sarif_input:
        slither.sarif_input = args.sarif_input
    if args.sarif_triage:
        slither.sarif_triage = args.sarif_triage

    return _process(slither, detector_classes, printer_classes)


def process_all(
    target: str,
    args: argparse.Namespace,
    detector_classes: list[type[AbstractDetector]],
    printer_classes: list[type[AbstractPrinter]],
) -> tuple[list[Slither], list[dict], list[Output], int]:
    compilations = compile_all(target, **vars(args))
    slither_instances = []
    results_detectors = []
    results_printers = []
    analyzed_contracts_count = 0
    for compilation in compilations:
        (
            slither,
            current_results_detectors,
            current_results_printers,
            current_analyzed_count,
        ) = process_single(compilation, args, detector_classes, printer_classes)
        results_detectors.extend(current_results_detectors)
        results_printers.extend(current_results_printers)
        slither_instances.append(slither)
        analyzed_contracts_count += current_analyzed_count
    return (
        slither_instances,
        results_detectors,
        results_printers,
        analyzed_contracts_count,
    )


def _process(
    slither: Slither,
    detector_classes: list[type[AbstractDetector]],
    printer_classes: list[type[AbstractPrinter]],
) -> tuple[Slither, list[dict], list[Output], int]:
    for detector_cls in detector_classes:
        slither.register_detector(detector_cls)

    for printer_cls in printer_classes:
        slither.register_printer(printer_cls)

    analyzed_contracts_count = len(slither.contracts)

    results_detectors = []
    results_printers = []

    if not printer_classes:
        detector_resultss = slither.run_detectors()
        detector_resultss = [x for x in detector_resultss if x]  # remove empty results
        detector_results = [item for sublist in detector_resultss for item in sublist]  # flatten
        results_detectors.extend(detector_results)

    else:
        printer_results = slither.run_printers()
        printer_results = [x for x in printer_results if x]  # remove empty results
        results_printers.extend(printer_results)

    return slither, results_detectors, results_printers, analyzed_contracts_count


# endregion
###################################################################################
###################################################################################

# region Detectors and printers
###################################################################################
###################################################################################


def get_detectors_and_printers() -> tuple[
    list[type[AbstractDetector]], list[type[AbstractPrinter]]
]:
    detectors_ = [getattr(all_detectors, name) for name in dir(all_detectors)]
    detectors = [d for d in detectors_ if inspect.isclass(d) and issubclass(d, AbstractDetector)]

    printers_ = [getattr(all_printers, name) for name in dir(all_printers)]
    printers = [p for p in printers_ if inspect.isclass(p) and issubclass(p, AbstractPrinter)]

    # Handle plugins!
    entry_points = metadata.entry_points(group="slither_analyzer.plugin")

    for entry_point in entry_points:
        make_plugin = entry_point.load()

        plugin_detectors, plugin_printers = make_plugin()

        detector = None
        if not all(issubclass(detector, AbstractDetector) for detector in plugin_detectors):
            raise ValueError(
                f"Error when loading plugin {entry_point}, {detector} is not a detector"
            )
        printer = None
        if not all(issubclass(printer, AbstractPrinter) for printer in plugin_printers):
            raise ValueError(f"Error when loading plugin {entry_point}, {printer} is not a printer")

        # We convert those to lists in case someone returns a tuple
        detectors += list(plugin_detectors)
        printers += list(plugin_printers)

    return detectors, printers


def choose_detectors(
    args: argparse.Namespace, all_detector_classes: list[type[AbstractDetector]]
) -> list[type[AbstractDetector]]:
    # If detectors are specified, run only these ones

    detectors_to_run = []
    detectors = {d.ARGUMENT: d for d in all_detector_classes}

    if args.detectors_to_run == "all":
        detectors_to_run = all_detector_classes
    else:
        detectors_to_run = __include_detectors(
            set(detectors_to_run), args.detectors_to_run, detectors
        )
        return detectors_to_run

    classification_map = {
        DetectorClassification.HIGH: args.exclude_high,
        DetectorClassification.MEDIUM: args.exclude_medium,
        DetectorClassification.LOW: args.exclude_low,
        DetectorClassification.INFORMATIONAL: args.exclude_informational,
        DetectorClassification.OPTIMIZATION: args.exclude_optimization,
    }
    excluded_classification = [
        classification for classification, included in classification_map.items() if included
    ]
    detectors_to_run = [d for d in detectors_to_run if d.IMPACT not in excluded_classification]

    if args.detectors_to_exclude:
        detectors_to_run = [
            d for d in detectors_to_run if d.ARGUMENT not in args.detectors_to_exclude
        ]

    if args.detectors_to_include:
        detectors_to_run = __include_detectors(
            set(detectors_to_run), args.detectors_to_include, detectors
        )

    detectors_to_run = sorted(detectors_to_run, key=lambda x: x.IMPACT)
    return detectors_to_run


def __include_detectors(
    detectors_to_run: set[type[AbstractDetector]],
    detectors_to_include: str,
    detectors: dict[str, type[AbstractDetector]],
) -> list[type[AbstractDetector]]:
    include_detectors = detectors_to_include.split(",")

    for detector in include_detectors:
        if detector in detectors:
            detectors_to_run.add(detectors[detector])
        else:
            raise ValueError(f"Error: {detector} is not a detector")

    return detectors_to_run


def choose_printers(
    args: argparse.Namespace, all_printer_classes: list[type[AbstractPrinter]]
) -> list[type[AbstractPrinter]]:
    printers_to_run = []

    # disable default printer
    if args.printers_to_run is None:
        return []

    if args.printers_to_run == "all":
        return all_printer_classes

    printers = {p.ARGUMENT: p for p in all_printer_classes}
    for printer in args.printers_to_run.split(","):
        if printer in printers:
            printers_to_run.append(printers[printer])
        else:
            raise ValueError(f"Error: {printer} is not a printer")
    return printers_to_run


# endregion
###################################################################################
###################################################################################
# region Command line parsing
###################################################################################
###################################################################################


def parse_filter_paths(args: argparse.Namespace, filter_path: bool) -> list[str]:
    paths = args.filter_paths if filter_path else args.include_paths
    if paths:
        return paths.split(",")
    return []


def parse_args(
    detector_classes: list[type[AbstractDetector]], printer_classes: list[type[AbstractPrinter]]
) -> argparse.Namespace:
    usage = "slither target [flag]\n"
    usage += "\ntarget can be:\n"
    usage += "\t- file.sol // a Solidity file\n"
    usage += "\t- project_directory // a project directory. See https://github.com/crytic/crytic-compile/#crytic-compile for the supported platforms\n"
    usage += "\t- 0x.. // a contract on mainnet\n"
    usage += f"\t- NETWORK:0x.. // a contract on a different network. Supported networks: {','.join(x for x in SUPPORTED_NETWORK)}\n"

    parser = argparse.ArgumentParser(
        description="For usage information, see https://github.com/crytic/slither/wiki/Usage",
        usage=usage,
    )

    parser.add_argument("filename", help=argparse.SUPPRESS)

    cryticparser.init(parser)

    parser.add_argument(
        "--version",
        help="displays the current version",
        version=metadata.version("slither-analyzer"),
        action="version",
    )

    group_detector = parser.add_argument_group("Detectors")
    group_printer = parser.add_argument_group("Printers")
    group_checklist = parser.add_argument_group(
        "Checklist (consider using https://github.com/crytic/slither-action)"
    )
    group_misc = parser.add_argument_group("Additional options")
    group_filters = parser.add_mutually_exclusive_group()

    group_detector.add_argument(
        "--detect",
        help="Comma-separated list of detectors, defaults to all, "
        f"available detectors: {', '.join(d.ARGUMENT for d in detector_classes)}",
        action="store",
        dest="detectors_to_run",
        default=defaults_flag_in_config["detectors_to_run"],
    )

    group_printer.add_argument(
        "--print",
        help="Comma-separated list of contract information printers, "
        f"available printers: {', '.join(d.ARGUMENT for d in printer_classes)}",
        action="store",
        dest="printers_to_run",
        default=defaults_flag_in_config["printers_to_run"],
    )

    group_printer.add_argument(
        "--include-interfaces",
        help="Include interfaces from inheritance-graph printer",
        action="store_true",
        dest="include_interfaces",
        default=False,
    )

    group_detector.add_argument(
        "--list-detectors",
        help="List available detectors",
        action=ListDetectors,
        nargs=0,
        default=False,
    )

    group_printer.add_argument(
        "--list-printers",
        help="List available printers",
        action=ListPrinters,
        nargs=0,
        default=False,
    )

    group_detector.add_argument(
        "--exclude",
        help="Comma-separated list of detectors that should be excluded",
        action="store",
        dest="detectors_to_exclude",
        default=defaults_flag_in_config["detectors_to_exclude"],
    )

    group_detector.add_argument(
        "--exclude-dependencies",
        help="Exclude results that are only related to dependencies",
        action="store_true",
        default=defaults_flag_in_config["exclude_dependencies"],
    )

    group_detector.add_argument(
        "--exclude-optimization",
        help="Exclude optimization analyses",
        action="store_true",
        default=defaults_flag_in_config["exclude_optimization"],
    )

    group_detector.add_argument(
        "--exclude-informational",
        help="Exclude informational impact analyses",
        action="store_true",
        default=defaults_flag_in_config["exclude_informational"],
    )

    group_detector.add_argument(
        "--exclude-low",
        help="Exclude low impact analyses",
        action="store_true",
        default=defaults_flag_in_config["exclude_low"],
    )

    group_detector.add_argument(
        "--exclude-medium",
        help="Exclude medium impact analyses",
        action="store_true",
        default=defaults_flag_in_config["exclude_medium"],
    )

    group_detector.add_argument(
        "--exclude-high",
        help="Exclude high impact analyses",
        action="store_true",
        default=defaults_flag_in_config["exclude_high"],
    )

    group_detector.add_argument(
        "--include-detectors",
        help="Comma-separated list of detectors that should be included",
        action="store",
        dest="detectors_to_include",
        default=defaults_flag_in_config["detectors_to_include"],
    )

    fail_on_group = group_detector.add_mutually_exclusive_group()
    fail_on_group.add_argument(
        "--fail-pedantic",
        help="Fail if any findings are detected",
        action="store_const",
        dest="fail_on",
        const=FailOnLevel.PEDANTIC,
    )
    fail_on_group.add_argument(
        "--fail-low",
        help="Fail if any low or greater impact findings are detected",
        action="store_const",
        dest="fail_on",
        const=FailOnLevel.LOW,
    )
    fail_on_group.add_argument(
        "--fail-medium",
        help="Fail if any medium or greater impact findings are detected",
        action="store_const",
        dest="fail_on",
        const=FailOnLevel.MEDIUM,
    )
    fail_on_group.add_argument(
        "--fail-high",
        help="Fail if any high impact findings are detected",
        action="store_const",
        dest="fail_on",
        const=FailOnLevel.HIGH,
    )
    fail_on_group.add_argument(
        "--fail-none",
        "--no-fail-pedantic",
        help="Do not return the number of findings in the exit code",
        action="store_const",
        dest="fail_on",
        const=FailOnLevel.NONE,
    )
    fail_on_group.set_defaults(fail_on=FailOnLevel.PEDANTIC)

    group_detector.add_argument(
        "--show-ignored-findings",
        help="Show all the findings",
        action="store_true",
        default=defaults_flag_in_config["show_ignored_findings"],
    )

    group_checklist.add_argument(
        "--checklist",
        help="Generate a markdown page with the detector results",
        action="store_true",
        default=False,
    )

    group_checklist.add_argument(
        "--checklist-limit",
        help="Limit the number of results per detector in the markdown file",
        action="store",
        default="",
    )

    group_checklist.add_argument(
        "--markdown-root",
        type=check_and_sanitize_markdown_root,
        help="URL for markdown generation",
        action="store",
        default="",
    )

    group_misc.add_argument(
        "--json",
        help='Export the results as a JSON file ("--json -" to export to stdout)',
        action="store",
        default=defaults_flag_in_config["json"],
    )

    group_misc.add_argument(
        "--sarif",
        help='Export the results as a SARIF JSON file ("--sarif -" to export to stdout)',
        action="store",
        default=defaults_flag_in_config["sarif"],
    )

    group_misc.add_argument(
        "--sarif-input",
        help="Sarif input (beta)",
        action="store",
        default=defaults_flag_in_config["sarif_input"],
    )

    group_misc.add_argument(
        "--sarif-triage",
        help="Sarif triage (beta)",
        action="store",
        default=defaults_flag_in_config["sarif_triage"],
    )

    group_misc.add_argument(
        "--json-types",
        help="Comma-separated list of result types to output to JSON, defaults to "
        + f"{','.join(output_type for output_type in DEFAULT_JSON_OUTPUT_TYPES)}. "
        + f"Available types: {','.join(output_type for output_type in JSON_OUTPUT_TYPES)}",
        action="store",
        default=defaults_flag_in_config["json-types"],
    )

    group_misc.add_argument(
        "--zip",
        help="Export the results as a zipped JSON file",
        action="store",
        default=defaults_flag_in_config["zip"],
    )

    group_misc.add_argument(
        "--zip-type",
        help=f"Zip compression type. One of {','.join(ZIP_TYPES_ACCEPTED.keys())}. Default lzma",
        action="store",
        default=defaults_flag_in_config["zip_type"],
    )

    group_misc.add_argument(
        "--disable-color",
        help="Disable output colorization",
        action="store_true",
        default=defaults_flag_in_config["disable_color"],
    )

    group_misc.add_argument(
        "--triage-mode",
        help="Run triage mode (save results in triage database)",
        action="store_true",
        dest="triage_mode",
        default=False,
    )

    group_misc.add_argument(
        "--triage-database",
        help="File path to the triage database (default: slither.db.json)",
        action="store",
        dest="triage_database",
        default=defaults_flag_in_config["triage_database"],
    )

    group_misc.add_argument(
        "--config-file",
        help="Provide a config file (default: slither.config.json)",
        action="store",
        dest="config_file",
        default=None,
    )

    group_misc.add_argument(
        "--change-line-prefix",
        help="Change the line prefix (default #) for the displayed source codes (i.e. file.sol#1).",
        action="store",
        dest="change_line_prefix",
        default="#",
    )

    group_misc.add_argument(
        "--solc-ast",
        help="Provide the contract as a json AST",
        action="store_true",
        default=False,
    )

    group_misc.add_argument(
        "--generate-patches",
        help="Generate patches (json output only)",
        action="store_true",
        default=False,
    )

    group_misc.add_argument(
        "--no-fail",
        help="Do not fail in case of parsing (echidna mode only)",
        action="store_true",
        default=defaults_flag_in_config["no_fail"],
    )

    group_filters.add_argument(
        "--filter-paths",
        help="Regex filter to exclude detector results matching file path e.g. (mocks/|test/)",
        action="store",
        dest="filter_paths",
        default=defaults_flag_in_config["filter_paths"],
    )

    group_filters.add_argument(
        "--include-paths",
        help="Regex filter to include detector results matching file path e.g. (src/|contracts/). Opposite of --filter-paths",
        action="store",
        dest="include_paths",
        default=defaults_flag_in_config["include_paths"],
    )

    # debugger command
    parser.add_argument("--debug", help=argparse.SUPPRESS, action="store_true", default=False)

    parser.add_argument("--markdown", help=argparse.SUPPRESS, action=OutputMarkdown, default=False)

    parser.add_argument(
        "--wiki-detectors", help=argparse.SUPPRESS, action=OutputWiki, default=False
    )

    parser.add_argument(
        "--list-detectors-json",
        help=argparse.SUPPRESS,
        action=ListDetectorsJson,
        nargs=0,
        default=False,
    )

    parser.add_argument(
        "--legacy-ast",
        help=argparse.SUPPRESS,
        action="store_true",
        default=defaults_flag_in_config["legacy_ast"],
    )

    parser.add_argument(
        "--skip-assembly",
        help=argparse.SUPPRESS,
        action="store_true",
        default=defaults_flag_in_config["skip_assembly"],
    )

    parser.add_argument(
        "--perf",
        help=argparse.SUPPRESS,
        action="store_true",
        default=False,
    )

    # Disable the throw/catch on partial analyses
    parser.add_argument(
        "--disallow-partial", help=argparse.SUPPRESS, action="store_true", default=False
    )

    if len(sys.argv) == 1:
        parser.print_help(sys.stderr)
        sys.exit(1)

    args = parser.parse_args()
    read_config_file(args)

    args.filter_paths = parse_filter_paths(args, True)
    args.include_paths = parse_filter_paths(args, False)

    # Verify our json-type output is valid
    args.json_types = set(args.json_types.split(","))  # type:ignore
    for json_type in args.json_types:
        if json_type not in JSON_OUTPUT_TYPES:
            raise ValueError(f'Error: "{json_type}" is not a valid JSON result output type.')

    return args


class ListDetectors(argparse.Action):
    def __call__(self, parser: Any, *args: Any, **kwargs: Any) -> None:
        detectors, _ = get_detectors_and_printers()
        output_detectors(detectors)
        parser.exit()


class ListDetectorsJson(argparse.Action):
    def __call__(self, parser: Any, *args: Any, **kwargs: Any) -> None:
        detectors, _ = get_detectors_and_printers()
        detector_types_json = output_detectors_json(detectors)
        print(json.dumps(detector_types_json))
        parser.exit()


class ListPrinters(argparse.Action):
    def __call__(self, parser: Any, *args: Any, **kwargs: Any) -> None:
        _, printers = get_detectors_and_printers()
        output_printers(printers)
        parser.exit()


class OutputMarkdown(argparse.Action):
    def __call__(
        self,
        parser: Any,
        args: Any,
        values: str | Sequence[Any] | None,
        option_string: Any = None,
    ) -> None:
        detectors, printers = get_detectors_and_printers()
        assert isinstance(values, str)
        output_to_markdown(detectors, printers, values)
        parser.exit()


class OutputWiki(argparse.Action):
    def __call__(
        self,
        parser: Any,
        args: Any,
        values: str | Sequence[Any] | None,
        option_string: Any = None,
    ) -> None:
        detectors, _ = get_detectors_and_printers()
        assert isinstance(values, str)
        output_wiki(detectors, values)
        parser.exit()


# endregion
###################################################################################
###################################################################################
# region CustomFormatter
###################################################################################
###################################################################################


class FormatterCryticCompile(logging.Formatter):
    def format(self, record: logging.LogRecord) -> str:
        # for i, msg in enumerate(record.msg):
        if record.msg.startswith("Compilation warnings/errors on "):
            txt = record.args[1]  # type:ignore
            txt = txt.split("\n")  # type:ignore
            txt = [red(x) if "Error" in x else x for x in txt]
            txt = "\n".join(txt)
            record.args = (record.args[0], txt)  # type:ignore
        return super().format(record)


# endregion
###################################################################################
###################################################################################
# region Main
###################################################################################
###################################################################################


def main() -> None:
    # Codebase with complex domninators can lead to a lot of SSA recursive call
    sys.setrecursionlimit(1500)

    detectors, printers = get_detectors_and_printers()

    main_impl(all_detector_classes=detectors, all_printer_classes=printers)


def main_impl(
    all_detector_classes: list[type[AbstractDetector]],
    all_printer_classes: list[type[AbstractPrinter]],
) -> None:
    """
    :param all_detector_classes: A list of all detectors that can be included/excluded.
    :param all_printer_classes: A list of all printers that can be included.
    """
    # Set logger of Slither to info, to catch warnings related to the arg parsing
    logger.setLevel(logging.INFO)
    args = parse_args(all_detector_classes, all_printer_classes)

    cp: cProfile.Profile | None = None
    if args.perf:
        cp = cProfile.Profile()
        cp.enable()

    # Set colorization option
    set_colorization_enabled(False if args.disable_color else sys.stdout.isatty())

    # Define some variables for potential JSON output
    json_results: dict[str, Any] = {}
    output_error = None
    outputting_json = args.json is not None
    outputting_json_stdout = args.json == "-"
    outputting_sarif = args.sarif is not None
    outputting_sarif_stdout = args.sarif == "-"
    outputting_zip = args.zip is not None
    if args.zip_type not in ZIP_TYPES_ACCEPTED:
        to_log = f"Zip type not accepted, it must be one of {','.join(ZIP_TYPES_ACCEPTED.keys())}"
        logger.error(to_log)

    # If we are outputting JSON, capture all standard output. If we are outputting to stdout, we block typical stdout
    # output.
    if outputting_json or outputting_sarif:
        StandardOutputCapture.enable(outputting_json_stdout or outputting_sarif_stdout)

    printer_classes = choose_printers(args, all_printer_classes)
    detector_classes = choose_detectors(args, all_detector_classes)

    default_log = logging.INFO if not args.debug else logging.DEBUG

    for l_name, l_level in [
        ("Slither", default_log),
        ("Contract", default_log),
        ("Function", default_log),
        ("Node", default_log),
        ("Parsing", default_log),
        ("Detectors", default_log),
        ("FunctionSolc", default_log),
        ("ExpressionParsing", default_log),
        ("TypeParsing", default_log),
        ("SSA_Conversion", default_log),
        ("Printers", default_log),
        # ('CryticCompile', default_log)
    ]:
        logger_level = logging.getLogger(l_name)
        logger_level.setLevel(l_level)

    console_handler = logging.StreamHandler()
    console_handler.setLevel(logging.INFO)

    console_handler.setFormatter(FormatterCryticCompile())

    crytic_compile_error = logging.getLogger("CryticCompile")
    crytic_compile_error.addHandler(console_handler)
    crytic_compile_error.propagate = False
    crytic_compile_error.setLevel(logging.INFO)

    results_detectors: list[dict] = []
    results_printers: list[Output] = []
    try:
        filename = args.filename

        # Determine if we are handling ast from solc
        if args.solc_ast or (filename.endswith(".json") and not is_supported(filename)):
            globbed_filenames = glob.glob(filename, recursive=True)
            filenames = glob.glob(os.path.join(filename, "*.json"))
            if not filenames:
                filenames = globbed_filenames
            number_contracts = 0

            slither_instances = []
            for filename in filenames:
                (
                    slither_instance,
                    results_detectors_tmp,
                    results_printers_tmp,
                    number_contracts_tmp,
                ) = process_single(filename, args, detector_classes, printer_classes)
                number_contracts += number_contracts_tmp
                results_detectors += results_detectors_tmp
                results_printers += results_printers_tmp
                slither_instances.append(slither_instance)

        # Rely on CryticCompile to discern the underlying type of compilations.
        else:
            (
                slither_instances,
                results_detectors,
                results_printers,
                number_contracts,
            ) = process_all(filename, args, detector_classes, printer_classes)

        # Determine if we are outputting JSON
        if outputting_json or outputting_zip or output_to_sarif:
            # Add our compilation information to JSON
            if "compilations" in args.json_types:
                compilation_results = []
                for slither_instance in slither_instances:
                    assert slither_instance.crytic_compile
                    compilation_results.append(
                        generate_standard_export(slither_instance.crytic_compile)
                    )
                json_results["compilations"] = compilation_results

            # Add our detector results to JSON if desired.
            if results_detectors and "detectors" in args.json_types:
                json_results["detectors"] = results_detectors

            # Add our printer results to JSON if desired.
            if results_printers and "printers" in args.json_types:
                json_results["printers"] = results_printers

            # Add our detector types to JSON
            if "list-detectors" in args.json_types:
                detectors, _ = get_detectors_and_printers()
                json_results["list-detectors"] = output_detectors_json(detectors)

            # Add our detector types to JSON
            if "list-printers" in args.json_types:
                _, printers = get_detectors_and_printers()
                json_results["list-printers"] = output_printers_json(printers)

        # Output our results to markdown if we wish to compile a checklist.
        if args.checklist:
            output_results_to_markdown(
                results_detectors, args.checklist_limit, args.show_ignored_findings
            )

        # Don't print the number of result for printers
        if number_contracts == 0:
            logger.warning(red("No contract was analyzed"))
        if printer_classes:
            logger.info("%s analyzed (%d contracts)", filename, number_contracts)
        else:
            logger.info(
                "%s analyzed (%d contracts with %d detectors), %d result(s) found",
                filename,
                number_contracts,
                len(detector_classes),
                len(results_detectors),
            )

    except SlitherException as slither_exception:
        output_error = str(slither_exception)
        traceback.print_exc()
        logging.error(red("Error:"))
        logging.error(red(output_error))
        logging.error("Please report an issue to https://github.com/crytic/slither/issues")

    # If we are outputting JSON, capture the redirected output and disable the redirect to output the final JSON.
    if outputting_json:
        if "console" in args.json_types:
            json_results["console"] = {
                "stdout": StandardOutputCapture.get_stdout_output(),
                "stderr": StandardOutputCapture.get_stderr_output(),
            }
        StandardOutputCapture.disable()
        output_to_json(None if outputting_json_stdout else args.json, output_error, json_results)

    if outputting_sarif:
        StandardOutputCapture.disable()
        output_to_sarif(
            None if outputting_sarif_stdout else args.sarif, json_results, detector_classes
        )

    if outputting_zip:
        output_to_zip(args.zip, output_error, json_results, args.zip_type)

    if args.perf and cp:
        cp.disable()
        stats = pstats.Stats(cp).sort_stats("cumtime")
        stats.print_stats()

    fail_on = FailOnLevel(args.fail_on)
    if fail_on == FailOnLevel.HIGH:
        fail_on_detection = any(result["impact"] == "High" for result in results_detectors)
    elif fail_on == FailOnLevel.MEDIUM:
        fail_on_detection = any(
            result["impact"] in ["Medium", "High"] for result in results_detectors
        )
    elif fail_on == FailOnLevel.LOW:
        fail_on_detection = any(
            result["impact"] in ["Low", "Medium", "High"] for result in results_detectors
        )
    elif fail_on == FailOnLevel.PEDANTIC:
        fail_on_detection = bool(results_detectors)
    else:
        fail_on_detection = False

    # Exit with them appropriate status code
    if output_error or fail_on_detection:
        sys.exit(-1)
    else:
        sys.exit(0)


if __name__ == "__main__":
    main()

# endregion



"""
Module detecting modifiers that are not guaranteed to execute _; or revert()/throw

Note that require()/assert() are not considered here. Even if they
are in the outermost scope, they do not guarantee a revert, so a
default value can still be returned.
"""

from slither.detectors.abstract_detector import (
    AbstractDetector,
    DetectorClassification,
    DETECTOR_INFO,
)
from slither.core.cfg.node import Node, NodeType
from slither.utils.output import Output


def is_revert(node: Node) -> bool:
    return node.type == NodeType.THROW or any(
        ir.function.name in ["revert()", "revert(string"] for ir in node.internal_calls
    )


def _get_false_son(node: Node) -> Node:
    """Select the son node corresponding to a false branch
    Following this node stays on the outer scope of the function
    """
    if node.type == NodeType.IF:
        return node.sons[1]

    if node.type == NodeType.IFLOOP:
        return next(s for s in node.sons if s.type == NodeType.ENDLOOP)

    return None


class ModifierDefaultDetection(AbstractDetector):
    """
    Detector for modifiers that return a default value
    """

    ARGUMENT = "incorrect-modifier"
    HELP = "Modifiers that can return the default value"
    IMPACT = DetectorClassification.LOW
    CONFIDENCE = DetectorClassification.HIGH
    WIKI = "https://github.com/crytic/slither/wiki/Detector-Documentation#incorrect-modifier"

    WIKI_TITLE = "Incorrect modifier"
    WIKI_DESCRIPTION = "If a modifier does not execute `_` or revert, the execution of the function will return the default value, which can be misleading for the caller."

    # region wiki_exploit_scenario
    WIKI_EXPLOIT_SCENARIO = """
```solidity
    modidfier myModif(){
        if(..){
           _;
        }
    }
    function get() myModif returns(uint){

    }
```
If the condition in `myModif` is false, the execution of `get()` will return 0."""
    # endregion wiki_exploit_scenario

    WIKI_RECOMMENDATION = "All the paths in a modifier must execute `_` or revert."

    def _detect(self) -> list[Output]:
        results = []

        for c in self.contracts:
            for mod in c.modifiers:
                if mod.contract_declarer != c:
                    continue
                # Walk down the tree, only looking at nodes in the outer scope
                node = mod.entry_point
                node_seen = []
                while node is not None:
                    node_seen.append(node)
                    # If any node in the outer scope executes _; or reverts,
                    # we will never return a default value
                    if node.type == NodeType.PLACEHOLDER or is_revert(node):
                        break

                    # Move down, staying on the outer scope in branches
                    if len(node.sons) > 0:
                        if node.contains_if():
                            node = _get_false_son(node)
                        else:
                            if node.sons[0] in node_seen:
                                node = node.sons[1]
                            else:
                                node = node.sons[0]
                    else:
                        node = None
                else:
                    # Nothing was found in the outer scope
                    info: DETECTOR_INFO = [
                        "Modifier ",
                        mod,
                        " does not always execute _; or revert\n",
                    ]

                    res = self.generate_result(info)
                    results.append(res)

        return results



from pathlib import Path

from slither import Slither
from slither.slithir.operations.internal_call import InternalCall
from slither.slithir.operations.new_structure import NewStructure
from slither.slithir.variables.constant import Constant

TEST_DATA_DIR = Path(__file__).resolve().parent / "test_data"
ARG_REORDER_TEST_ROOT = Path(TEST_DATA_DIR, "argument_reorder")


def test_struct_constructor_reorder(solc_binary_path) -> None:
    solc_path = solc_binary_path("0.8.15")
    slither = Slither(
        Path(ARG_REORDER_TEST_ROOT, "test_struct_constructor.sol").as_posix(), solc=solc_path
    )

    operations = slither.contracts[0].functions[0].slithir_operations
    constructor_calls = [x for x in operations if isinstance(x, NewStructure)]
    assert len(constructor_calls) == 2

    # Arguments to first call are 2, 3
    assert (
        isinstance(constructor_calls[0].arguments[0], Constant)
        and constructor_calls[0].arguments[0].value == 2
    )
    assert (
        isinstance(constructor_calls[0].arguments[1], Constant)
        and constructor_calls[0].arguments[1].value == 3
    )

    # Arguments to second call are 5, 4 (note the reversed order)
    assert (
        isinstance(constructor_calls[1].arguments[0], Constant)
        and constructor_calls[1].arguments[0].value == 5
    )
    assert (
        isinstance(constructor_calls[1].arguments[1], Constant)
        and constructor_calls[1].arguments[1].value == 4
    )


def test_internal_call_reorder(solc_binary_path) -> None:
    solc_path = solc_binary_path("0.8.15")
    slither = Slither(
        Path(ARG_REORDER_TEST_ROOT, "test_internal_call_reorder.sol").as_posix(), solc=solc_path
    )

    operations = slither.contracts[0].functions[1].slithir_operations
    internal_calls = [x for x in operations if isinstance(x, InternalCall)]
    assert len(internal_calls) == 1

    # Arguments to call are 3, true, 5
    assert (
        isinstance(internal_calls[0].arguments[0], Constant)
        and internal_calls[0].arguments[0].value == 3
    )
    assert (
        isinstance(internal_calls[0].arguments[1], Constant)
        and internal_calls[0].arguments[1].value is True
    )
    assert (
        isinstance(internal_calls[0].arguments[2], Constant)
        and internal_calls[0].arguments[2].value == 5
    )


def test_overridden_function_reorder(solc_binary_path) -> None:
    solc_path = solc_binary_path("0.8.15")
    slither = Slither(
        Path(ARG_REORDER_TEST_ROOT, "test_overridden_function.sol").as_posix(), solc=solc_path
    )

    operations = slither.contracts[0].functions[1].slithir_operations
    internal_calls = [x for x in operations if isinstance(x, InternalCall)]
    assert len(internal_calls) == 1

    assert (
        isinstance(internal_calls[0].arguments[0], Constant)
        and internal_calls[0].arguments[0].value == 23
    )
    assert (
        isinstance(internal_calls[0].arguments[1], Constant)
        and internal_calls[0].arguments[1].value == 36
    )
    assert (
        isinstance(internal_calls[0].arguments[2], Constant)
        and internal_calls[0].arguments[2].value == 34
    )

    operations = slither.contracts[1].functions[1].slithir_operations
    internal_calls = [x for x in operations if isinstance(x, InternalCall)]
    assert len(internal_calls) == 1

    assert (
        isinstance(internal_calls[0].arguments[0], Constant)
        and internal_calls[0].arguments[0].value == 23
    )
    assert (
        isinstance(internal_calls[0].arguments[1], Constant)
        and internal_calls[0].arguments[1].value == 36
    )
    assert (
        isinstance(internal_calls[0].arguments[2], Constant)
        and internal_calls[0].arguments[2].value == 34
    )



from slither.slithir.operations.operation import Operation
from slither.slithir.utils.utils import is_valid_rvalue, RVALUE


class Condition(Operation):
    """
    Condition
    Only present as last operation in conditional node
    """

    def __init__(
        self,
        value: RVALUE,
    ) -> None:
        assert is_valid_rvalue(value)
        super().__init__()
        self._value = value

    @property
    def read(
        self,
    ) -> list[RVALUE]:
        return [self.value]

    @property
    def value(self) -> RVALUE:
        return self._value

    def __str__(self) -> str:
        return f"CONDITION {self.value}"



import re
import shutil
from collections import Counter
from pathlib import Path
import pytest

from crytic_compile import CryticCompile
from crytic_compile.platform.solc_standard_json import SolcStandardJson

from slither import Slither
from slither.printers.inheritance.inheritance_graph import PrinterInheritanceGraph
from slither.printers.summary.cheatcodes import CheatcodePrinter
from slither.printers.summary.slithir import PrinterSlithIR


TEST_DATA_DIR = Path(__file__).resolve().parent / "test_data"

foundry_available = shutil.which("forge") is not None
project_ready = Path(TEST_DATA_DIR, "test_printer_cheatcode/lib/forge-std").exists()


def test_inheritance_printer(solc_binary_path) -> None:
    solc_path = solc_binary_path("0.8.0")
    standard_json = SolcStandardJson()
    standard_json.add_source_file(Path(TEST_DATA_DIR, "test_contract_names", "A.sol").as_posix())
    standard_json.add_source_file(Path(TEST_DATA_DIR, "test_contract_names", "B.sol").as_posix())
    standard_json.add_source_file(Path(TEST_DATA_DIR, "test_contract_names", "B2.sol").as_posix())
    standard_json.add_source_file(Path(TEST_DATA_DIR, "test_contract_names", "C.sol").as_posix())
    compilation = CryticCompile(standard_json, solc=solc_path)
    slither = Slither(compilation)
    printer = PrinterInheritanceGraph(slither=slither, logger=None)

    output = printer.output("test_printer.dot")
    content = output.elements[0]["name"]["content"]

    pattern = re.compile(r"(?:c\d+_)?(\w+ -> )(?:c\d+_)(\w+)")
    matches = re.findall(pattern, content)
    relations = ["".join(m) for m in matches]

    counter = Counter(relations)

    assert counter["B -> A"] == 2
    assert counter["C -> A"] == 1
    # Let also test the include/exclude interface behavior
    # Check that the interface is not included
    assert "MyInterfaceX" not in content

    slither.include_interfaces = True
    output = printer.output("test_printer.dot")
    content = output.elements[0]["name"]["content"]
    assert "MyInterfaceX" in content

    # Remove test generated files
    Path("test_printer.dot").unlink(missing_ok=True)


@pytest.mark.skipif(
    not foundry_available or not project_ready, reason="requires Foundry and project setup"
)
def test_printer_cheatcode():
    slither = Slither(
        Path(TEST_DATA_DIR, "test_printer_cheatcode").as_posix(), foundry_compile_all=True
    )

    printer = CheatcodePrinter(slither=slither, logger=None)
    output = printer.output("")

    assert (
        output.data["description"]
        == "CounterTest (test/Counter.t.sol)\n\tsetUp\n\t\tdeal - (test/Counter.t.sol#21 (9 - 32)\n\t\tvm.deal(alice,1000000000000000000)\n\n\t\tdeal - (test/Counter.t.sol#22 (9 - 30)\n\t\tvm.deal(bob,2000000000000000000)\n\n\ttestIncrement\n\t\tprank - (test/Counter.t.sol#28 (9 - 24)\n\t\tvm.prank(alice)\n\n\t\tassertEq - (test/Counter.t.sol#30 (9 - 38)\n\t\tassertEq(counter.number(),1)\n\n\t\tprank - (test/Counter.t.sol#32 (9 - 22)\n\t\tvm.prank(bob)\n\n\t\tassertEq - (test/Counter.t.sol#34 (9 - 38)\n\t\tassertEq(counter.number(),2)\n\n"
    )


def test_slithir_printer(solc_binary_path) -> None:
    solc_path = solc_binary_path("0.8.0")
    standard_json = SolcStandardJson()
    standard_json.add_source_file(
        Path(TEST_DATA_DIR, "test_printer_slithir", "bug-2266.sol").as_posix()
    )
    compilation = CryticCompile(standard_json, solc=solc_path)
    slither = Slither(compilation)

    printer = PrinterSlithIR(slither, logger=None)
    output = printer.output("test_printer_slithir.dot")

    assert "slither.core.solidity_types" not in output.data["description"]



"""
Module detecting possible loss of precision due to divide before multiple
"""

from collections import defaultdict

from slither.core.cfg.node import Node
from slither.core.declarations.contract import Contract
from slither.core.declarations.function_contract import FunctionContract
from slither.detectors.abstract_detector import (
    AbstractDetector,
    DetectorClassification,
    DETECTOR_INFO,
)
from slither.slithir.operations import Binary, Assignment, BinaryType, LibraryCall, Operation
from slither.slithir.utils.utils import LVALUE
from slither.slithir.variables import Constant
from slither.utils.output import Output


def is_division(ir: Operation) -> bool:
    if isinstance(ir, Binary):
        if ir.type == BinaryType.DIVISION:
            return True

    if isinstance(ir, LibraryCall):
        if ir.function.name and ir.function.name.lower() in [
            "div",
            "safediv",
        ]:
            if len(ir.arguments) == 2:
                if ir.lvalue:
                    return True
    return False


def is_multiplication(ir: Operation) -> bool:
    if isinstance(ir, Binary):
        if ir.type == BinaryType.MULTIPLICATION:
            return True

    if isinstance(ir, LibraryCall):
        if ir.function.name and ir.function.name.lower() in [
            "mul",
            "safemul",
        ]:
            if len(ir.arguments) == 2:
                if ir.lvalue:
                    return True
    return False


def is_assert(node: Node) -> bool:
    if node.contains_require_or_assert():
        return True
    # Old Solidity code where using an internal 'assert(bool)' function
    # While we dont check that this function is correct, we assume it is
    # To avoid too many FP
    if "assert(bool)" in [ir.function.full_name for ir in node.internal_calls]:
        return True
    return False


def _explore(
    to_explore: list[Node], f_results: list[list[Node]], divisions: defaultdict[LVALUE, list[Node]]
) -> None:
    explored = set()
    while to_explore:
        node = to_explore.pop()

        if node in explored:
            continue
        explored.add(node)

        equality_found = False
        # List of nodes related to one bug instance
        node_results: list[Node] = []

        for ir in node.irs:
            if isinstance(ir, Assignment):
                if ir.rvalue in divisions:
                    # Avoid duplicate. We dont use set so we keep the order of the nodes
                    if node not in divisions[ir.rvalue]:  # type: ignore
                        divisions[ir.lvalue] = divisions[ir.rvalue] + [node]  # type: ignore
                    else:
                        divisions[ir.lvalue] = divisions[ir.rvalue]  # type: ignore

            if is_division(ir):
                divisions[ir.lvalue] = [node]  # type: ignore

            if is_multiplication(ir):
                mul_arguments = ir.read if isinstance(ir, Binary) else ir.arguments  # type: ignore
                nodes = []
                for r in mul_arguments:
                    if not isinstance(r, Constant) and (r in divisions):
                        # Dont add node already present to avoid duplicate
                        # We dont use set to keep the order of the nodes
                        if node in divisions[r]:
                            nodes += [n for n in divisions[r] if n not in nodes]
                        else:
                            nodes += [n for n in divisions[r] + [node] if n not in nodes]
                if nodes:
                    node_results = nodes

            if isinstance(ir, Binary) and ir.type == BinaryType.EQUAL:
                equality_found = True

        if node_results:
            # We do not track the case where the multiplication is done in a require() or assert()
            # Which also contains a ==, to prevent FP due to the form
            # assert(a == b * c + a % b)
            if not (is_assert(node) and equality_found):
                f_results.append(node_results)

        for son in node.sons:
            to_explore.append(son)


def detect_divide_before_multiply(
    contract: Contract,
) -> list[tuple[FunctionContract, list[Node]]]:
    """
    Detects and returns all nodes with multiplications of division results.
    :param contract: Contract to detect assignment within.
    :return: A list of nodes with multiplications of divisions.
    """

    # Create our result set.
    # List of tuple (function -> list(list(nodes)))
    # Each list(nodes) of the list is one bug instances
    # Each node in the list(nodes) is involved in the bug
    results: list[tuple[FunctionContract, list[Node]]] = []

    # Loop for each function and modifier.
    for function in contract.functions_declared + contract.modifiers_declared:
        if not function.entry_point:
            continue

        # List of list(nodes)
        # Each list(nodes) is one bug instances
        f_results: list[list[Node]] = []

        # lvalue -> node
        # track all the division results (and the assignment of the division results)
        divisions: defaultdict[LVALUE, list[Node]] = defaultdict(list)

        _explore([function.entry_point], f_results, divisions)

        for f_result in f_results:
            results.append((function, f_result))

    # Return the resulting set of nodes with divisions before multiplications
    return results


class DivideBeforeMultiply(AbstractDetector):
    """
    Divide before multiply
    """

    ARGUMENT = "divide-before-multiply"
    HELP = "Imprecise arithmetic operations order"
    IMPACT = DetectorClassification.MEDIUM
    CONFIDENCE = DetectorClassification.MEDIUM

    WIKI = "https://github.com/crytic/slither/wiki/Detector-Documentation#divide-before-multiply"

    WIKI_TITLE = "Divide before multiply"
    WIKI_DESCRIPTION = """Solidity's integer division truncates. Thus, performing division before multiplication can lead to precision loss."""

    # region wiki_exploit_scenario
    WIKI_EXPLOIT_SCENARIO = """
```solidity
contract A {
	function f(uint n) public {
        coins = (oldSupply / n) * interest;
    }
}
```
If `n` is greater than `oldSupply`, `coins` will be zero. For example, with `oldSupply = 5; n = 10, interest = 2`, coins will be zero.
If `(oldSupply * interest / n)` was used, `coins` would have been `1`.
In general, it's usually a good idea to re-arrange arithmetic to perform multiplication before division, unless the limit of a smaller type makes this dangerous."""
    # endregion wiki_exploit_scenario

    WIKI_RECOMMENDATION = """Consider ordering multiplication before division."""

    def _detect(self) -> list[Output]:
        """
        Detect divisions before multiplications
        """
        results = []
        for contract in self.contracts:
            divisions_before_multiplications = detect_divide_before_multiply(contract)
            if divisions_before_multiplications:
                for func, nodes in divisions_before_multiplications:
                    info: DETECTOR_INFO = [
                        func,
                        " performs a multiplication on the result of a division:\n",
                    ]

                    # sort the nodes to get deterministic results
                    nodes.sort(key=lambda x: x.node_id)

                    for node in nodes:
                        info += ["\t- ", node, "\n"]

                    res = self.generate_result(info)
                    results.append(res)

        return results



"""
Detect events with address parameters that have no indexed parameters
"""

from slither.core.declarations.event import Event
from slither.core.solidity_types.elementary_type import ElementaryType
from slither.detectors.abstract_detector import AbstractDetector, DetectorClassification
from slither.utils.output import Output


class UnindexedEventAddress(AbstractDetector):
    """
    Detect events with address parameters that have no indexed parameters.

    Indexing address parameters enables efficient off-chain filtering of events.
    """

    ARGUMENT = "unindexed-event-address"
    HELP = "Events with address parameters but no indexed parameters"
    IMPACT = DetectorClassification.INFORMATIONAL
    CONFIDENCE = DetectorClassification.HIGH

    WIKI = (
        "https://github.com/crytic/slither/wiki/Detector-Documentation"
        "#unindexed-event-address-parameters"
    )

    WIKI_TITLE = "Unindexed event address parameters"
    WIKI_DESCRIPTION = (
        "Detects events that have address-type parameters but no indexed parameters. "
        "Indexing event parameters enables efficient off-chain filtering."
    )

    WIKI_EXPLOIT_SCENARIO = """
```solidity
contract Token {
    event Transfer(address from, address to, uint256 value);

    function transfer(address to, uint256 value) external {
        // ...
        emit Transfer(msg.sender, to, value);
    }
}
```
The `Transfer` event has address parameters but none are indexed.
Off-chain tools cannot efficiently filter transfers by sender or recipient address."""

    WIKI_RECOMMENDATION = (
        "Add the `indexed` keyword to address parameters in events "
        "to enable efficient off-chain filtering."
    )

    @staticmethod
    def _has_unindexed_address(event: Event) -> bool:
        """
        Check if an event has address parameters but no indexed parameters.

        Args:
            event: The event to check.

        Returns:
            True if the event has address params but no indexed params.
        """
        has_address = False
        has_indexed = False

        for param in event.elems:
            if param.indexed:
                has_indexed = True
            if param.type == ElementaryType("address"):
                has_address = True

        return has_address and not has_indexed

    def _detect(self) -> list[Output]:
        """
        Detect events with address parameters but no indexed parameters.

        Returns:
            List of detection results.
        """
        results: list[Output] = []

        # Check contract-level events
        for contract in self.contracts:
            for event in contract.events_declared:
                if self._has_unindexed_address(event):
                    info = [
                        "Event ",
                        event,
                        " has address parameters but no indexed parameters\n",
                    ]
                    res = self.generate_result(info)
                    results.append(res)

        # Check top-level events
        for event in self.compilation_unit.events_top_level:
            if self._has_unindexed_address(event):
                info = [
                    "Event ",
                    event,
                    " has address parameters but no indexed parameters\n",
                ]
                res = self.generate_result(info)
                results.append(res)

        return results



"""
Compute the data depenency between all the SSA variables
"""

from collections import defaultdict
from typing import TYPE_CHECKING

from slither.core.cfg.node import Node
from slither.core.declarations import (
    Contract,
    Enum,
    Function,
    SolidityFunction,
    SolidityVariable,
    SolidityVariableComposed,
    Structure,
    FunctionContract,
)
from slither.core.declarations.solidity_import_placeholder import SolidityImportPlaceHolder
from slither.core.solidity_types.type import Type
from slither.core.variables.top_level_variable import TopLevelVariable
from slither.core.variables.variable import Variable
from slither.slithir.operations import Index, OperationWithLValue, InternalCall, Operation
from slither.slithir.utils.utils import LVALUE
from slither.slithir.variables import (
    Constant,
    LocalIRVariable,
    ReferenceVariable,
    ReferenceVariableSSA,
    StateIRVariable,
    TemporaryVariableSSA,
    TupleVariableSSA,
)
from slither.slithir.variables.variable import SlithIRVariable

if TYPE_CHECKING:
    from slither.core.compilation_unit import SlitherCompilationUnit

###################################################################################
###################################################################################
# region User APIs
###################################################################################
###################################################################################


SUPPORTED_TYPES = Variable | SolidityVariable

# TODO refactor the data deps to be better suited for top level function object
# Right now we allow to pass a node to ease the API, but we need something
# better
# The deps propagation for top level elements is also not working as expected
Context_types_API = Contract | Function | Node
Context_types = Contract | Function


def is_dependent(
    variable: SUPPORTED_TYPES,
    source: SUPPORTED_TYPES,
    context: Context_types_API,
    only_unprotected: bool = False,
) -> bool:
    """
    If Node is provided as context, the context will be the broader context, either the contract or the function,
    depending on if the node is in a top level function or not

    Args:
        variable (Variable)
        source (Variable)
        context (Contract|Function|Node).
        only_unprotected (bool): True only unprotected function are considered
    Returns:
        bool
    """
    assert isinstance(context, (Contract, Function, Node))
    if isinstance(context, Node):
        func = context.function
        context = func.contract if isinstance(func, FunctionContract) else func

    if isinstance(variable, Constant):
        return False
    if variable == source:
        return True
    context_dict = context.context

    if only_unprotected:
        return (
            variable in context_dict[KEY_NON_SSA_UNPROTECTED]
            and source in context_dict[KEY_NON_SSA_UNPROTECTED][variable]
        )
    return variable in context_dict[KEY_NON_SSA] and source in context_dict[KEY_NON_SSA][variable]


def is_dependent_ssa(
    variable: SUPPORTED_TYPES,
    source: SUPPORTED_TYPES,
    context: Context_types_API,
    only_unprotected: bool = False,
) -> bool:
    """
    If Node is provided as context, the context will be the broader context, either the contract or the function,
    depending on if the node is in a top level function or not

    Args:
        variable (Variable)
        taint (Variable)
        context (Contract|Function)
        only_unprotected (bool): True only unprotected function are considered
    Returns:
        bool
    """
    assert isinstance(context, (Contract, Function, Node))
    if isinstance(context, Node):
        func = context.function
        context = func.contract if isinstance(func, FunctionContract) else func
    context_dict = context.context
    if isinstance(variable, Constant):
        return False
    if variable == source:
        return True
    if only_unprotected:
        return (
            variable in context_dict[KEY_SSA_UNPROTECTED]
            and source in context_dict[KEY_SSA_UNPROTECTED][variable]
        )
    return variable in context_dict[KEY_SSA] and source in context_dict[KEY_SSA][variable]


GENERIC_TAINT = {
    SolidityVariableComposed("msg.sender"),
    SolidityVariableComposed("msg.value"),
    SolidityVariableComposed("msg.data"),
    SolidityVariableComposed("tx.origin"),
    SolidityVariableComposed("tx.gasprice"),
}


def is_tainted(
    variable: SUPPORTED_TYPES,
    context: Context_types_API,
    only_unprotected: bool = False,
    ignore_generic_taint: bool = False,
) -> bool:
    """
    If Node is provided as context, the context will be the broader context, either the contract or the function,
    depending on if the node is in a top level function or not

        Args:
        variable
        context (Contract|Function)
        only_unprotected (bool): True only unprotected function are considered
    Returns:
        bool
    """
    assert isinstance(context, (Contract, Function, Node))
    if isinstance(context, Node):
        func = context.function
        context = func.contract if isinstance(func, FunctionContract) else func
    assert isinstance(only_unprotected, bool)
    if isinstance(variable, Constant):
        return False
    compilation_unit = context.compilation_unit
    taints = compilation_unit.context[KEY_INPUT]
    if not ignore_generic_taint:
        taints |= GENERIC_TAINT
    return variable in taints or any(
        is_dependent(variable, t, context, only_unprotected) for t in taints
    )


def is_tainted_ssa(
    variable: SUPPORTED_TYPES,
    context: Context_types_API,
    only_unprotected: bool = False,
    ignore_generic_taint: bool = False,
) -> bool:
    """
    If Node is provided as context, the context will be the broader context, either the contract or the function,
    depending on if the node is in a top level function or not

    Args:
        variable
        context (Contract|Function)
        only_unprotected (bool): True only unprotected function are considered
    Returns:
        bool
    """
    assert isinstance(context, (Contract, Function, Node))
    if isinstance(context, Node):
        func = context.function
        context = func.contract if isinstance(func, FunctionContract) else func
    assert isinstance(only_unprotected, bool)
    if isinstance(variable, Constant):
        return False
    compilation_unit = context.compilation_unit
    taints = compilation_unit.context[KEY_INPUT_SSA]
    if not ignore_generic_taint:
        taints |= GENERIC_TAINT
    return variable in taints or any(
        is_dependent_ssa(variable, t, context, only_unprotected) for t in taints
    )


def get_dependencies(
    variable: SUPPORTED_TYPES,
    context: Context_types_API,
    only_unprotected: bool = False,
) -> set[Variable]:
    """
    Return the variables for which `variable` depends on.
    If Node is provided as context, the context will be the broader context, either the contract or the function,
    depending on if the node is in a top level function or not

    :param variable: The target
    :param context: Either a function (interprocedural) or a contract (inter transactional)
    :param only_unprotected: True if consider only protected functions
    :return: set(Variable)
    """
    assert isinstance(context, (Contract, Function, Node))
    if isinstance(context, Node):
        func = context.function
        context = func.contract if isinstance(func, FunctionContract) else func
    assert isinstance(only_unprotected, bool)
    if only_unprotected:
        return context.context[KEY_NON_SSA_UNPROTECTED].get(variable, set())
    return context.context[KEY_NON_SSA].get(variable, set())


def get_all_dependencies(
    context: Context_types_API, only_unprotected: bool = False
) -> dict[Variable, set[Variable]]:
    """
    Return the dictionary of dependencies.
    If Node is provided as context, the context will be the broader context, either the contract or the function,
    depending on if the node is in a top level function or not

    :param context: Either a function (interprocedural) or a contract (inter transactional)
    :param only_unprotected: True if consider only protected functions
    :return: Dict(Variable, set(Variable))
    """
    assert isinstance(context, (Contract, Function, Node))
    if isinstance(context, Node):
        func = context.function
        context = func.contract if isinstance(func, FunctionContract) else func
    assert isinstance(only_unprotected, bool)
    if only_unprotected:
        return context.context[KEY_NON_SSA_UNPROTECTED]
    return context.context[KEY_NON_SSA]


def get_dependencies_ssa(
    variable: SUPPORTED_TYPES,
    context: Context_types_API,
    only_unprotected: bool = False,
) -> set[Variable]:
    """
    Return the variables for which `variable` depends on (SSA version).
    If Node is provided as context, the context will be the broader context, either the contract or the function,
    depending on if the node is in a top level function or not

    :param variable: The target (must be SSA variable)
    :param context: Either a function (interprocedural) or a contract (inter transactional)
    :param only_unprotected: True if consider only protected functions
    :return: set(Variable)
    """
    assert isinstance(context, (Contract, Function, Node))
    if isinstance(context, Node):
        func = context.function
        context = func.contract if isinstance(func, FunctionContract) else func
    assert isinstance(only_unprotected, bool)
    if only_unprotected:
        return context.context[KEY_SSA_UNPROTECTED].get(variable, set())
    return context.context[KEY_SSA].get(variable, set())


def get_all_dependencies_ssa(
    context: Context_types_API, only_unprotected: bool = False
) -> dict[Variable, set[Variable]]:
    """
    Return the dictionary of dependencies.
    If Node is provided as context, the context will be the broader context, either the contract or the function,
    depending on if the node is in a top level function or not

    :param context: Either a function (interprocedural) or a contract (inter transactional)
    :param only_unprotected: True if consider only protected functions
    :return: Dict(Variable, set(Variable))
    """
    assert isinstance(context, (Contract, Function, Node))
    if isinstance(context, Node):
        func = context.function
        context = func.contract if isinstance(func, FunctionContract) else func
    assert isinstance(only_unprotected, bool)
    if only_unprotected:
        return context.context[KEY_SSA_UNPROTECTED]
    return context.context[KEY_SSA]


# endregion
###################################################################################
###################################################################################
# region Module constants
###################################################################################
###################################################################################

KEY_SSA = "DATA_DEPENDENCY_SSA"
KEY_NON_SSA = "DATA_DEPENDENCY"

# Only for unprotected functions
KEY_SSA_UNPROTECTED = "DATA_DEPENDENCY_SSA_UNPROTECTED"
KEY_NON_SSA_UNPROTECTED = "DATA_DEPENDENCY_UNPROTECTED"

KEY_INPUT = "DATA_DEPENDENCY_INPUT"
KEY_INPUT_SSA = "DATA_DEPENDENCY_INPUT_SSA"


# endregion
###################################################################################
###################################################################################
# region Debug
###################################################################################
###################################################################################


def pprint_dependency(caller_context: Context_types) -> None:
    print("#### SSA ####")
    context = caller_context.context
    for k, values in context[KEY_SSA].items():
        print(f"{k} ({id(k)}):")
        for v in values:
            print(f"\t- {v}")

    print("#### NON SSA ####")
    for k, values in context[KEY_NON_SSA].items():
        print(f"{k} ({hex(id(k))}):")
        for v in values:
            print(f"\t- {v} ({hex(id(v))})")


# endregion
###################################################################################
###################################################################################
# region Analyses
###################################################################################
###################################################################################


def compute_dependency(compilation_unit: "SlitherCompilationUnit") -> None:
    compilation_unit.context[KEY_INPUT] = set()
    compilation_unit.context[KEY_INPUT_SSA] = set()

    for contract in compilation_unit.contracts:
        compute_dependency_contract(contract, compilation_unit)


def compute_dependency_contract(
    contract: Contract, compilation_unit: "SlitherCompilationUnit"
) -> None:
    if KEY_SSA in contract.context:
        return

    contract.context[KEY_SSA] = {}
    contract.context[KEY_SSA_UNPROTECTED] = {}

    for function in contract.functions + list(contract.modifiers):
        compute_dependency_function(function)

        propagate_function(contract, function, KEY_SSA, KEY_NON_SSA)
        propagate_function(contract, function, KEY_SSA_UNPROTECTED, KEY_NON_SSA_UNPROTECTED)

        if function.visibility in ["public", "external"]:
            [compilation_unit.context[KEY_INPUT].add(p) for p in function.parameters]
            [compilation_unit.context[KEY_INPUT_SSA].add(p) for p in function.parameters_ssa]

    propagate_contract(contract, KEY_SSA, KEY_NON_SSA)
    propagate_contract(contract, KEY_SSA_UNPROTECTED, KEY_NON_SSA_UNPROTECTED)


def propagate_function(
    contract: Contract, function: Function, context_key: str, context_key_non_ssa: str
) -> None:
    transitive_close_dependencies(function, context_key, context_key_non_ssa)
    # Propage data dependency
    data_depencencies = function.context[context_key]
    for key, values in data_depencencies.items():
        if key not in contract.context[context_key]:
            contract.context[context_key][key] = set(values)
        else:
            contract.context[context_key][key].union(values)


def transitive_close_dependencies(
    context: Context_types, context_key: str, context_key_non_ssa: str
) -> None:
    # transitive closure
    changed = True
    keys = context.context[context_key].keys()
    while changed:
        changed = False
        to_add = defaultdict(set)
        for key, items in context.context[context_key].items():
            for item in items & keys:
                to_add[key].update(context.context[context_key][item] - {key} - items)
        for k, v in to_add.items():
            # Because we dont have any check on the update operation
            # We might update an empty set with an empty set
            if v:
                changed = True
                context.context[context_key][k] |= v
    context.context[context_key_non_ssa] = convert_to_non_ssa(context.context[context_key])


def propagate_contract(contract: Contract, context_key: str, context_key_non_ssa: str) -> None:
    transitive_close_dependencies(contract, context_key, context_key_non_ssa)


def add_dependency(lvalue: Variable, function: Function, ir: Operation, is_protected: bool) -> None:
    if lvalue not in function.context[KEY_SSA]:
        function.context[KEY_SSA][lvalue] = set()
        if not is_protected:
            function.context[KEY_SSA_UNPROTECTED][lvalue] = set()
    read: list[LVALUE | SolidityVariableComposed] | list[SlithIRVariable]
    if isinstance(ir, Index):
        read = [ir.variable_left]
    elif isinstance(ir, InternalCall) and ir.function:
        read = ir.function.return_values_ssa
    else:
        read = ir.read
    for v in read:
        if not isinstance(v, Constant):
            function.context[KEY_SSA][lvalue].add(v)
    if not is_protected:
        for v in read:
            if not isinstance(v, Constant):
                function.context[KEY_SSA_UNPROTECTED][lvalue].add(v)


def compute_dependency_function(function: Function) -> None:
    if KEY_SSA in function.context:
        return

    function.context[KEY_SSA] = {}
    function.context[KEY_SSA_UNPROTECTED] = {}

    is_protected = function.is_protected()
    for node in function.nodes:
        for ir in node.irs_ssa:
            if isinstance(ir, OperationWithLValue) and ir.lvalue:
                if isinstance(ir.lvalue, LocalIRVariable) and ir.lvalue.is_storage:
                    continue
                if isinstance(ir.lvalue, ReferenceVariable):
                    lvalue = ir.lvalue.points_to
                    if lvalue:
                        add_dependency(lvalue, function, ir, is_protected)
                add_dependency(ir.lvalue, function, ir, is_protected)

    function.context[KEY_NON_SSA] = convert_to_non_ssa(function.context[KEY_SSA])
    function.context[KEY_NON_SSA_UNPROTECTED] = convert_to_non_ssa(
        function.context[KEY_SSA_UNPROTECTED]
    )


def convert_variable_to_non_ssa(v: SUPPORTED_TYPES) -> SUPPORTED_TYPES:
    if isinstance(
        v,
        (
            LocalIRVariable,
            StateIRVariable,
            TemporaryVariableSSA,
            ReferenceVariableSSA,
            TupleVariableSSA,
        ),
    ):
        return v.non_ssa_version
    assert isinstance(
        v,
        (
            Constant,
            SolidityVariable,
            Contract,
            Enum,
            SolidityFunction,
            Structure,
            Function,
            Type,
            SolidityImportPlaceHolder,
            TopLevelVariable,
        ),
    )
    return v


def convert_to_non_ssa(
    data_depencies: dict[SUPPORTED_TYPES, set[SUPPORTED_TYPES]],
) -> dict[SUPPORTED_TYPES, set[SUPPORTED_TYPES]]:
    # Need to create new set() as its changed during iteration
    ret: dict[SUPPORTED_TYPES, set[SUPPORTED_TYPES]] = {}
    for k, values in data_depencies.items():
        var = convert_variable_to_non_ssa(k)
        if var not in ret:
            ret[var] = set()
        ret[var] = ret[var].union({convert_variable_to_non_ssa(v) for v in values})

    return ret



#
import pathlib
from argparse import ArgumentTypeError
from collections import defaultdict
from inspect import getsourcefile
from collections.abc import Callable

import pytest
from solc_select.solc_select import valid_version as solc_valid_version

from slither import Slither
from slither.core.cfg.node import Node, NodeType
from slither.core.declarations import Function, Contract
from slither.core.solidity_types import ArrayType, ElementaryType
from slither.core.variables.local_variable import LocalVariable
from slither.core.variables.state_variable import StateVariable
from slither.slithir.operations import (
    OperationWithLValue,
    Phi,
    Assignment,
    HighLevelCall,
    Return,
    Operation,
    Binary,
    BinaryType,
    InternalCall,
    Index,
    InitArray,
    NewArray,
)
from slither.slithir.utils.ssa import is_used_later
from slither.slithir.variables import (
    Constant,
    ReferenceVariable,
    LocalIRVariable,
    StateIRVariable,
    TemporaryVariableSSA,
)

# Directory of currently executing script. Will be used as basis for temporary file names.
SCRIPT_DIR = pathlib.Path(getsourcefile(lambda: 0)).parent  # type:ignore


def valid_version(ver: str) -> bool:
    """Wrapper function to check if the solc-version is valid

    The solc_select function raises and exception but for checks below,
    only a bool is needed.
    """
    try:
        solc_valid_version(ver)
        return True
    except ArgumentTypeError:
        return False


def have_ssa_if_ir(function: Function) -> None:
    """Verifies that all nodes in a function that have IR also have SSA IR"""
    for n in function.nodes:
        if n.irs:
            assert n.irs_ssa


def ssa_basic_properties(function: Function) -> None:
    """Verifies that basic properties of ssa holds

    1. Every name is defined only once
    2. A l-value is never index zero - there is always a zero-value available for each var
    3. Every r-value is at least defined at some point
    4. The number of ssa defs is >= the number of assignments to var
    5. Function parameters SSA are stored in function.parameters_ssa
       - if function parameter is_storage it refers to a fake variable
    6. Function returns SSA are stored in function.returns_ssa
        - if function return is_storage it refers to a fake variable
    """
    ssa_lvalues = set()
    ssa_rvalues = set()
    lvalue_assignments: dict[str, int] = {}

    for n in function.nodes:
        for ir in n.irs:
            if isinstance(ir, OperationWithLValue) and ir.lvalue:
                name = ir.lvalue.name
                if name is None:
                    continue
                if name in lvalue_assignments:
                    lvalue_assignments[name] += 1
                else:
                    lvalue_assignments[name] = 1

        for ssa in n.irs_ssa:
            if isinstance(ssa, OperationWithLValue):
                # 1
                assert ssa.lvalue not in ssa_lvalues
                ssa_lvalues.add(ssa.lvalue)

                # 2 (if Local/State Var)
                ssa_lvalue = ssa.lvalue
                if isinstance(ssa_lvalue, (StateIRVariable, LocalIRVariable)):
                    assert ssa_lvalue.index > 0

            for rvalue in filter(
                lambda x: not isinstance(x, (StateIRVariable, Constant)), ssa.read
            ):
                ssa_rvalues.add(rvalue)

    # 3
    # Each var can have one non-defined value, the value initially held. Typically,
    # var_0, i_0, state_0 or similar.
    undef_vars = set()
    for rvalue in ssa_rvalues:
        if rvalue not in ssa_lvalues:
            assert rvalue.non_ssa_version not in undef_vars
            undef_vars.add(rvalue.non_ssa_version)

    # 4
    ssa_defs: dict[str, int] = defaultdict(int)
    for v in ssa_lvalues:
        if v and v.name:
            ssa_defs[v.name] += 1

    for k, count in lvalue_assignments.items():
        assert ssa_defs[k] >= count

    # Helper 5/6
    def check_property_5_and_6(
        variables: list[LocalVariable], ssavars: list[LocalIRVariable]
    ) -> None:
        for var in filter(lambda x: x.name, variables):
            ssa_vars = [x for x in ssavars if x.non_ssa_version == var]
            assert len(ssa_vars) == 1
            ssa_var = ssa_vars[0]
            assert var.is_storage == ssa_var.is_storage
            if ssa_var.is_storage:
                assert len(ssa_var.refers_to) == 1
                assert ssa_var.refers_to[0].location == "reference_to_storage"

    # 5
    check_property_5_and_6(function.parameters, function.parameters_ssa)

    # 6
    check_property_5_and_6(function.returns, function.returns_ssa)


def ssa_phi_node_properties(f: Function) -> None:
    """Every phi-function should have as many args as predecessors

    This does not apply if the phi-node refers to state variables,
    they make use os special phi-nodes for tracking potential values
    a state variable can have
    """
    for node in f.nodes:
        for ssa in node.irs_ssa:
            if isinstance(ssa, Phi):
                n = len(ssa.read)
                if not isinstance(ssa.lvalue, StateIRVariable):
                    assert len(node.fathers) == n


# TODO (hbrodin): This should probably go into another file, not specific to SSA
def dominance_properties(f: Function) -> None:
    """Verifies properties related to dominators holds

    1. Every node have an immediate dominator except entry_node which have none
    2. From every node immediate dominator there is a path via its successors to the node
    """

    def find_path(from_node: Node, to: Node) -> bool:
        visited = set()
        worklist = list(from_node.sons)
        while worklist:
            first, *worklist = worklist
            if first == to:
                return True
            visited.add(first)
            for successor in first.sons:
                if successor not in visited:
                    worklist.append(successor)
        return False

    for node in f.nodes:
        if node is f.entry_point:
            assert node.immediate_dominator is None
        else:
            assert node.immediate_dominator is not None
            assert find_path(node.immediate_dominator, node)


def phi_values_inserted(f: Function) -> None:
    """Verifies that phi-values are inserted at the right places

    For every node that has a dominance frontier, any def (including
    phi) should be a phi function in its dominance frontier
    """

    def have_phi_for_var(
        node: Node, var: StateIRVariable | LocalIRVariable | TemporaryVariableSSA
    ) -> bool:
        """Checks if a node has a phi-instruction for var

        The ssa version would ideally be checked, but then
        more data flow analysis would be needed, for cases
        where a new def for var is introduced before reaching
        DF
        """
        non_ssa = var.non_ssa_version
        for ssa in node.irs_ssa:
            if isinstance(ssa, Phi):
                if non_ssa in (
                    r.non_ssa_version
                    for r in ssa.read
                    if isinstance(r, (StateIRVariable, LocalIRVariable, TemporaryVariableSSA))
                ):
                    return True
        return False

    for node in filter(lambda n: n.dominance_frontier, f.nodes):
        for df in node.dominance_frontier:
            for ssa in node.irs_ssa:
                if isinstance(ssa, OperationWithLValue):
                    ssa_lvalue = ssa.lvalue
                    if isinstance(
                        ssa_lvalue, (StateIRVariable, LocalIRVariable, TemporaryVariableSSA)
                    ) and is_used_later(node, ssa_lvalue):
                        assert have_phi_for_var(df, ssa_lvalue)


def verify_properties_hold(slither: Slither) -> None:
    """Ensures that basic properties of SSA hold true"""

    def verify_func(func: Function) -> None:
        have_ssa_if_ir(func)
        phi_values_inserted(func)
        ssa_basic_properties(func)
        ssa_phi_node_properties(func)
        dominance_properties(func)

    def verify(slither: Slither) -> None:
        for cu in slither.compilation_units:
            for func in cu.functions_and_modifiers:
                _dump_function(func)
                verify_func(func)
            for contract in cu.contracts:
                for f in contract.functions:
                    if f.is_constructor or f.is_constructor_variables:
                        _dump_function(f)
                        verify_func(f)

    assert isinstance(slither, Slither)
    verify(slither)


def _dump_function(f: Function) -> None:
    """Helper function to print nodes/ssa ir for a function or modifier"""
    print(f"---- {f.name} ----")
    for n in f.nodes:
        print(n)
        for ir in n.irs_ssa:
            print(f"\t{ir}")
    print("")


def _dump_functions(c: Contract) -> None:
    """Helper function to print functions and modifiers of a contract"""
    for f in c.functions_and_modifiers:
        _dump_function(f)


def get_filtered_ssa(f: Function | Node, flt: Callable) -> list[Operation]:
    """Returns a list of all ssanodes filtered by filter for all nodes in function f"""
    if isinstance(f, Function):
        return [ssanode for node in f.nodes for ssanode in node.irs_ssa if flt(ssanode)]

    assert isinstance(f, Node)
    return [ssanode for ssanode in f.irs_ssa if flt(ssanode)]


def get_ssa_of_type(f: Function | Node, ssatype) -> list[Operation]:
    """Returns a list of all ssanodes of a specific type for all nodes in function f"""
    return get_filtered_ssa(f, lambda ssanode: isinstance(ssanode, ssatype))


def test_multi_write(slither_from_solidity_source) -> None:
    source = """
    pragma solidity ^0.8.11;
    contract Test {
    function multi_write(uint val) external pure returns(uint) {
        val = 1;
        val = 2;
        val = 3;
    }
    }"""
    with slither_from_solidity_source(source) as slither:
        verify_properties_hold(slither)


def test_single_branch_phi(slither_from_solidity_source) -> None:
    source = """
        pragma solidity ^0.8.11;
        contract Test {
        function single_branch_phi(uint val) external pure returns(uint) {
            if (val == 3) {
                val = 9;
            }
            return val;
        }
        }
        """
    with slither_from_solidity_source(source) as slither:
        verify_properties_hold(slither)


def test_basic_phi(slither_from_solidity_source) -> None:
    source = """
    pragma solidity ^0.8.11;
    contract Test {
    function basic_phi(uint val) external pure returns(uint) {
        if (val == 3) {
            val = 9;
        } else {
            val = 1;
        }
        return val;
    }
    }
    """
    with slither_from_solidity_source(source) as slither:
        verify_properties_hold(slither)


def test_basic_loop_phi(slither_from_solidity_source) -> None:
    source = """
    pragma solidity ^0.8.11;
    contract Test {
    function basic_loop_phi(uint val) external pure returns(uint) {
        for (uint i=0;i<128;i++) {
            val = val + 1;
        }
        return val;
    }
    }
    """
    with slither_from_solidity_source(source) as slither:
        verify_properties_hold(slither)


@pytest.mark.xfail(strict=True, reason="Fails in current slither version. Fix in #1102.")
def test_phi_propagation_loop(slither_from_solidity_source):
    source = """
     pragma solidity ^0.8.11;
     contract Test {
     function looping(uint v) external pure returns(uint) {
        uint val = 0;
        for (uint i=0;i<v;i++) {
            if (val > i) {
                val = i;
            } else {
                val = 3;
            }
        }
        return val;
    }
    }
    """
    with slither_from_solidity_source(source) as slither:
        verify_properties_hold(slither)


@pytest.mark.xfail(strict=True, reason="Fails in current slither version. Fix in #1102.")
def test_free_function_properties(slither_from_solidity_source):
    source = """
        pragma solidity ^0.8.11;

        function free_looping(uint v) returns(uint) {
           uint val = 0;
           for (uint i=0;i<v;i++) {
               if (val > i) {
                   val = i;
               } else {
                   val = 3;
               }
           }
           return val;
       }

       contract Test {}
       """
    with slither_from_solidity_source(source) as slither:
        verify_properties_hold(slither)


def test_ssa_inter_transactional(slither_from_solidity_source) -> None:
    source = """
    pragma solidity ^0.8.11;
    contract A {
        uint my_var_A;
        uint my_var_B;

        function direct_set(uint i) public {
            my_var_A = i;
        }

        function direct_set_plus_one(uint i) public {
            my_var_A = i + 1;
        }

        function indirect_set() public {
            my_var_B = my_var_A;
        }
    }
    """
    with slither_from_solidity_source(source) as slither:
        c = slither.contracts[0]
        variables = c.variables_as_dict
        funcs = c.available_functions_as_dict()
        direct_set = funcs["direct_set(uint256)"]
        # Skip entry point and go straight to assignment ir
        assign1 = direct_set.nodes[1].irs_ssa[0]
        assert isinstance(assign1, Assignment)

        assign2 = direct_set.nodes[1].irs_ssa[0]
        assert isinstance(assign2, Assignment)

        indirect_set = funcs["indirect_set()"]
        phi = indirect_set.entry_point.irs_ssa[0]
        assert isinstance(phi, Phi)
        # phi rvalues come from 1, initial value of my_var_a and 2, assignment in direct_set
        assert len(phi.rvalues) == 3
        assert all(x.non_ssa_version == variables["my_var_A"] for x in phi.rvalues)
        assert assign1.lvalue in phi.rvalues
        assert assign2.lvalue in phi.rvalues


@pytest.mark.xfail(strict=True, reason="Fails in current slither version. Fix in #1102.")
def test_ssa_phi_callbacks(slither_from_solidity_source):
    source = """
    pragma solidity ^0.8.11;
    contract A {
        uint my_var_A;
        uint my_var_B;

        function direct_set(uint i) public {
            my_var_A = i;
        }

        function use_a() public {
            // Expect a phi-node here
            my_var_B = my_var_A;
            B b = new B();
            my_var_A = 3;
            b.do_stuff();
            // Expect a phi-node here
            my_var_B = my_var_A;
        }
    }

    contract B {
        function do_stuff() public returns (uint) {
            // This could be calling back into A
        }
    }
    """
    with slither_from_solidity_source(source) as slither:
        c = slither.get_contract_from_name("A")[0]
        _dump_functions(c)
        f = [x for x in c.functions if x.name == "use_a"][0]
        var_a = [x for x in c.variables if x.name == "my_var_A"][0]

        entry_phi = [
            x
            for x in f.entry_point.irs_ssa
            if isinstance(x, Phi) and x.lvalue.non_ssa_version == var_a
        ][0]
        # The four potential sources are:
        # 1. initial value
        # 2. my_var_A = i;
        # 3. my_var_A = 3;
        # 4. phi-value after call to b.do_stuff(), which could be reentrant.
        assert len(entry_phi.rvalues) == 4

        # Locate the first high-level call (should be b.do_stuff())
        call_node = [x for y in f.nodes for x in y.irs_ssa if isinstance(x, HighLevelCall)][0]
        n = call_node.node
        # Get phi-node after call
        after_call_phi = n.irs_ssa[n.irs_ssa.index(call_node) + 1]
        # The two sources for this phi node is
        # 1. my_var_A = i;
        # 2. my_var_A = 3;
        assert isinstance(after_call_phi, Phi)
        assert len(after_call_phi.rvalues) == 2


@pytest.mark.xfail(strict=True, reason="Fails in current slither version. Fix in #1102.")
def test_storage_refers_to(slither_from_solidity_source):
    """Test the storage aspects of the SSA IR

    When declaring a var as being storage, start tracking what storage it refers_to.
    When a phi-node is created, ensure refers_to is propagated to the phi-node.
    Assignments also propagate refers_to.
    Whenever a ReferenceVariable is the destination of an assignment (e.g. s.v = 10)
    below, create additional versions of the variables it refers to record that a
    write was made. In the current implementation, this is referenced by phis.
    """
    source = """
   contract A{

    struct St{
        int v;
    }

    St state0;
    St state1;

    function f() public{
        St storage s = state0;
        if(true){
            s = state1;
        }
        s.v = 10;
    }
}
    """
    with slither_from_solidity_source(source) as slither:
        c = slither.contracts[0]
        f = c.functions[0]

        phinodes = get_ssa_of_type(f, Phi)
        # Expect 2 in entrypoint (state0/state1 initial values), 1 at 'ENDIF' and two related to the
        # ReferenceVariable write s.v = 10.
        assert len(phinodes) == 5

        # Assign s to state0, s to state1, s.v to 10
        assigns = get_ssa_of_type(f, Assignment)
        assert len(assigns) == 3

        # The IR variables have is_storage
        assert all(x.lvalue.is_storage for x in assigns if isinstance(x, LocalIRVariable))

        # s.v ReferenceVariable points to one of the phi vars...
        ref0 = [x.lvalue for x in assigns if isinstance(x.lvalue, ReferenceVariable)][0]
        sphis = [x for x in phinodes if x.lvalue == ref0.points_to]
        assert len(sphis) == 1
        sphi = sphis[0]

        # ...and that phi refers to the two entry phi-values
        entryphi = [x for x in phinodes if x.lvalue in sphi.lvalue.refers_to]
        assert len(entryphi) == 2

        # The remaining two phis are the ones recording that write through ReferenceVariable occurred
        for ephi in entryphi:
            phinodes.remove(ephi)
        phinodes.remove(sphi)
        assert len(phinodes) == 2

        # And they are recorded in one of the entry phis
        assert phinodes[0].lvalue in entryphi[0].rvalues or entryphi[1].rvalues
        assert phinodes[1].lvalue in entryphi[0].rvalues or entryphi[1].rvalues


@pytest.mark.skipif(
    not valid_version("0.4.0"), reason="Solidity version 0.4.0 not available on this platform"
)
def test_initial_version_exists_for_locals(slither_from_solidity_source):
    """
    In solidity you can write statements such as
    uint a = a + 1, this test ensures that can be handled for local variables.
    """
    src = """
    contract C {
        function func() internal {
            uint a = a + 1;
        }
    }
    """
    with slither_from_solidity_source(src, "0.4.0") as slither:
        verify_properties_hold(slither)
        c = slither.contracts[0]
        f = c.functions[0]

        addition = get_ssa_of_type(f, Binary)[0]
        assert addition.type == BinaryType.ADDITION
        assert isinstance(addition.variable_right, Constant)
        a_0 = addition.variable_left
        assert a_0.index == 0
        assert a_0.name == "a"

        assignment = get_ssa_of_type(f, Assignment)[0]
        a_1 = assignment.lvalue
        assert a_1.index == 1
        assert a_1.name == "a"
        assert assignment.rvalue == addition.lvalue

        assert a_0.non_ssa_version == a_1.non_ssa_version


@pytest.mark.xfail(strict=True, reason="Fails in current slither version. Fix in #1102.")
@pytest.mark.skipif(
    not valid_version("0.4.0"), reason="Solidity version 0.4.0 not available on this platform"
)
def test_initial_version_exists_for_state_variables(slither_from_solidity_source):
    """
    In solidity you can write statements such as
    uint a = a + 1, this test ensures that can be handled for state variables.
    """
    src = """
    contract C {
        uint a = a + 1;
    }
    """
    with slither_from_solidity_source(src, "0.4.0") as slither:
        verify_properties_hold(slither)
        c = slither.contracts[0]
        f = c.functions[0]  # There will be one artificial ctor function for the state vars

        addition = get_ssa_of_type(f, Binary)[0]
        assert addition.type == BinaryType.ADDITION
        assert isinstance(addition.variable_right, Constant)
        a_0 = addition.variable_left
        assert isinstance(a_0, StateIRVariable)
        assert a_0.name == "a"

        assignment = get_ssa_of_type(f, Assignment)[0]
        a_1 = assignment.lvalue
        assert isinstance(a_1, StateIRVariable)
        assert a_1.index == a_0.index + 1
        assert a_1.name == "a"
        assert assignment.rvalue == addition.lvalue

        assert a_0.non_ssa_version == a_1.non_ssa_version
        assert isinstance(a_0.non_ssa_version, StateVariable)

        # No conditional/other function interaction so no phis
        assert len(get_ssa_of_type(f, Phi)) == 0


@pytest.mark.xfail(strict=True, reason="Fails in current slither version. Fix in #1102.")
def test_initial_version_exists_for_state_variables_function_assign(slither_from_solidity_source):
    """
    In solidity you can write statements such as
    uint a = a + 1, this test ensures that can be handled for local variables.
    """
    # TODO (hbrodin): Could be a detector that a is not used in f
    src = """
    contract C {
        uint a = f();

        function f() internal returns(uint) {
            return a;
        }
    }
    """
    with slither_from_solidity_source(src) as slither:
        verify_properties_hold(slither)
        c = slither.contracts[0]
        f, ctor = c.functions
        if f.is_constructor_variables:
            f, ctor = ctor, f

        # ctor should have a single call to f that assigns to a
        # temporary variable, that is then assigned to a

        call = get_ssa_of_type(ctor, InternalCall)[0]
        assert call.node.function == f
        assign = get_ssa_of_type(ctor, Assignment)[0]
        assert assign.rvalue == call.lvalue
        assert isinstance(assign.lvalue, StateIRVariable)
        assert assign.lvalue.name == "a"

        # f should have a phi node on entry of a0, a1 and should return
        # a2
        phi = get_ssa_of_type(f, Phi)[0]
        assert len(phi.rvalues) == 2
        assert assign.lvalue in phi.rvalues


@pytest.mark.skipif(
    not valid_version("0.4.0"), reason="Solidity version 0.4.0 not available on this platform"
)
def test_return_local_before_assign(slither_from_solidity_source):
    src = """
    // this require solidity < 0.5
    // a variable can be returned before declared. Ensure it can be
    // handled by Slither.
    contract A {
    function local(bool my_bool) internal returns(uint){
        if(my_bool){
            return a_local;
        }

        uint a_local = 10;
    }
    }
    """
    with slither_from_solidity_source(src, "0.4.0") as slither:
        f = slither.contracts[0].functions[0]

        ret = get_ssa_of_type(f, Return)[0]
        assert len(ret.values) == 1
        assert ret.values[0].index == 0

        assign = get_ssa_of_type(f, Assignment)[0]
        assert assign.lvalue.index == 1
        assert assign.lvalue.non_ssa_version == ret.values[0].non_ssa_version


@pytest.mark.skipif(
    not valid_version("0.5.0"), reason="Solidity version 0.5.0 not available on this platform"
)
def test_shadow_local(slither_from_solidity_source):
    src = """
    contract A {
     // this require solidity 0.5
    function shadowing_local() internal{
        uint local = 0;
        {
            uint local = 1;
            {
                uint local = 2;
            }
        }
    }
    }
    """
    with slither_from_solidity_source(src, "0.5.0") as slither:
        _dump_functions(slither.contracts[0])
        f = slither.contracts[0].functions[0]

        # Ensure all assignments are to a variable of index 1
        # not using the same IR var.
        assert all(x.lvalue.index == 1 for x in get_ssa_of_type(f, Assignment))


@pytest.mark.xfail(strict=True, reason="Fails in current slither version. Fix in #1102.")
def test_multiple_named_args_returns(slither_from_solidity_source):
    """Verifies that named arguments and return values have correct versions

    Each arg/ret have an initial version, version 0, and is written once and should
    then have version 1.
    """
    src = """
    contract A {
        function multi(uint arg1, uint arg2) internal returns (uint ret1, uint ret2) {
            arg1 = arg1 + 1;
            arg2 = arg2 + 2;
            ret1 = arg1 + 3;
            ret2 = arg2 + 4;
        }
    }"""
    with slither_from_solidity_source(src) as slither:
        verify_properties_hold(slither)
        f = slither.contracts[0].functions[0]

        # Ensure all LocalIRVariables (not TemporaryVariables) have index 1
        assert all(
            x.lvalue.index == 1 or not isinstance(x.lvalue, LocalIRVariable)
            for x in get_ssa_of_type(f, OperationWithLValue)
        )


@pytest.mark.xfail(reason="Tests for wanted state of SSA IR, not current.", strict=True)
def test_memory_array(slither_from_solidity_source):
    src = """
    contract MemArray {
        struct A {
            uint val1;
            uint val2;
        }

        function test_array() internal {
            A[] memory a= new A[](4);
            // Create REF_0 -> a_1[2]
            accept_array_entry(a[2]);

            // Create REF_1 -> a_1[3]
            accept_array_entry(a[3]);

            A memory alocal;
            accept_array_entry(alocal);

        }

        // val_1 = ϕ(val_0, REF_0, REF_1, alocal_1)
        // val_0 is an unknown external value
        function accept_array_entry(A memory val) public returns (uint) {
            uint zero = 0;
            b(zero);
            // Create REF_2 -> val_1.val1
            return b(val.val1);
        }

        function b(uint arg) public returns (uint){
            // arg_1 = ϕ(arg_0, zero_1, REF_2)
            return arg + 1;
        }
    }"""
    with slither_from_solidity_source(src) as slither:
        c = slither.contracts[0]

        ftest_array, faccept, fb = c.functions

        # Locate REF_0/REF_1/alocal (they are all args to the call)
        accept_args = [x.arguments[0] for x in get_ssa_of_type(ftest_array, InternalCall)]

        # Check entrypoint of accept_array_entry, it should contain a phi-node
        # of expected rvalues
        [phi_entry_accept] = get_ssa_of_type(faccept.entry_point, Phi)
        for arg in accept_args:
            assert arg in phi_entry_accept.rvalues
        # NOTE(hbrodin): There should be an additional val_0 in the phi-node.
        # That additional val_0 indicates an external caller of this function.
        assert len(phi_entry_accept.rvalues) == len(accept_args) + 1

        # Args used to invoke b
        b_args = [x.arguments[0] for x in get_ssa_of_type(faccept, InternalCall)]

        # Check entrypoint of B, it should contain a phi-node of expected
        # rvalues
        [phi_entry_b] = get_ssa_of_type(fb.entry_point, Phi)
        for arg in b_args:
            assert arg in phi_entry_b.rvalues

        # NOTE(hbrodin): There should be an additional arg_0 (see comment about phi_entry_accept).
        assert len(phi_entry_b.rvalues) == len(b_args) + 1


@pytest.mark.xfail(reason="Tests for wanted state of SSA IR, not current.", strict=True)
def test_storage_array(slither_from_solidity_source):
    src = """
    contract StorageArray {
        struct A {
            uint val1;
            uint val2;
        }

        // NOTE(hbrodin): a is never written, should only become a_0. Same for astorage (astorage_0). Phi-nodes at entry
        // should only add new versions of a state variable if it is actually written.
        A[] a;
        A astorage;

        function test_array() internal {
            accept_array_entry(a[2]);
            accept_array_entry(a[3]);
            accept_array_entry(astorage);
        }

        function accept_array_entry(A storage val) internal returns (uint) {
            // val is either a[2], a[3] or astorage_0. Ideally this could be identified.
            uint five = 5;

            // NOTE(hbrodin): If the following line is enabled, there would ideally be a phi-node representing writes
            // to either a or astorage.
            //val.val2 = 4;
            b(five);
            return b(val.val1);
        }

        function b(uint value) public returns (uint){
            // Expect a phi-node at the entrypoint
            // value_1 = ϕ(value_0, five_0, REF_x), where REF_x is the reference to val.val1 in accept_array_entry.
            return value + 1;
        }
    }"""
    with slither_from_solidity_source(src) as slither:
        c = slither.contracts[0]
        _dump_functions(c)
        ftest, faccept, fb = c.functions

        # None of a/astorage is written so expect that there are no phi-nodes at entrypoint.
        assert len(get_ssa_of_type(ftest.entry_point, Phi)) == 0

        # Expect all references to start from index 0 (no writes)
        assert all(x.variable_left.index == 0 for x in get_ssa_of_type(ftest, Index))

        [phi_entry_accept] = get_ssa_of_type(faccept.entry_point, Phi)
        assert len(phi_entry_accept.rvalues) == 3  # See comment in b above

        [phi_entry_b] = get_ssa_of_type(fb.entry_point, Phi)
        assert len(phi_entry_b.rvalues) == 3  # See comment in b above


@pytest.mark.xfail(strict=True, reason="Fails in current slither version. Fix in #1102.")
def test_issue_468(slither_from_solidity_source):
    """
    Ensure issue 468 is corrected as per
    https://github.com/crytic/slither/issues/468#issuecomment-620974151
    The one difference is that we allow the phi-function at entry of f to
    hold exit state which contains init state and state from branch, which
    is a bit redundant. This could be further simplified.
    """
    source = """
    contract State {
    int state = 0;
    function f(int a) public returns (int) {
        // phi-node here for state
        if (a < 1) {
            state += 1;
        }
        // phi-node here for state
        return state;
    }
    }
    """
    with slither_from_solidity_source(source) as slither:
        c = slither.get_contract_from_name("State")[0]
        f = [x for x in c.functions if x.name == "f"][0]

        # Check that there is an entry point phi values for each later value
        # plus one additional which is the initial value
        entry_ssa = f.entry_point.irs_ssa
        assert len(entry_ssa) == 1
        phi_entry = entry_ssa[0]
        assert isinstance(phi_entry, Phi)

        # Find the second phi function
        endif_node = [x for x in f.nodes if x.type == NodeType.ENDIF][0]
        assert len(endif_node.irs_ssa) == 1
        phi_endif = endif_node.irs_ssa[0]
        assert isinstance(phi_endif, Phi)

        # Ensure second phi-function contains init-phi and one additional
        assert len(phi_endif.rvalues) == 2
        assert phi_entry.lvalue in phi_endif.rvalues

        # Find return-statement and ensure it returns the phi_endif
        return_node = [x for x in f.nodes if x.type == NodeType.RETURN][0]
        assert len(return_node.irs_ssa) == 1
        ret = return_node.irs_ssa[0]
        assert len(ret.values) == 1
        assert phi_endif.lvalue in ret.values

        # Ensure that the phi_endif (which is the end-state for function as well) is in the entry_phi
        assert phi_endif.lvalue in phi_entry.rvalues


@pytest.mark.xfail(strict=True, reason="Fails in current slither version. Fix in #1102.")
def test_issue_434(slither_from_solidity_source):
    source = """
     contract Contract {
        int public a;
        function f() public {
            g();
            a += 1;
        }

        function e() public {
            a -= 1;
        }

        function g() public {
            e();
        }
    }
    """
    with slither_from_solidity_source(source) as slither:
        c = slither.get_contract_from_name("Contract")[0]

        e = [x for x in c.functions if x.name == "e"][0]
        f = [x for x in c.functions if x.name == "f"][0]
        g = [x for x in c.functions if x.name == "g"][0]

        # Ensure there is a phi-node at the beginning of f and e
        phi_entry_e = get_ssa_of_type(e.entry_point, Phi)[0]
        phi_entry_f = get_ssa_of_type(f.entry_point, Phi)[0]
        # But not at g
        assert len(get_ssa_of_type(g, Phi)) == 0

        # Ensure that the final states of f and e are in the entry-states
        add_f = get_filtered_ssa(
            f, lambda x: isinstance(x, Binary) and x.type == BinaryType.ADDITION
        )[0]
        sub_e = get_filtered_ssa(
            e, lambda x: isinstance(x, Binary) and x.type == BinaryType.SUBTRACTION
        )[0]
        assert add_f.lvalue in phi_entry_f.rvalues
        assert add_f.lvalue in phi_entry_e.rvalues
        assert sub_e.lvalue in phi_entry_f.rvalues
        assert sub_e.lvalue in phi_entry_e.rvalues

        # Ensure there is a phi-node after call to g
        call = get_ssa_of_type(f, InternalCall)[0]
        idx = call.node.irs_ssa.index(call)
        aftercall_phi = call.node.irs_ssa[idx + 1]
        assert isinstance(aftercall_phi, Phi)

        # Ensure that phi node ^ is used in the addition afterwards
        assert aftercall_phi.lvalue in (add_f.variable_left, add_f.variable_right)


@pytest.mark.xfail(strict=True, reason="Fails in current slither version. Fix in #1102.")
def test_issue_473(slither_from_solidity_source):
    source = """
    contract Contract {
    function f() public returns (int) {
        int a = 1;
        if (a > 0) {
            a = 2;
        }
        if (a == 3) {
            a = 6;
        }
        return a;
    }
    }
    """
    with slither_from_solidity_source(source) as slither:
        c = slither.get_contract_from_name("Contract")[0]
        f = c.functions[0]

        phis = get_ssa_of_type(f, Phi)
        return_value = get_ssa_of_type(f, Return)[0]

        # There shall be two phi functions
        assert len(phis) == 2
        first_phi = phis[0]
        second_phi = phis[1]

        # The second phi is the one being returned, if it's the first swap them (iteration order)
        if first_phi.lvalue in return_value.values:
            first_phi, second_phi = second_phi, first_phi

        # First phi is for [a=1 or a=2]
        assert len(first_phi.rvalues) == 2

        # second is for [a=6 or first phi]
        assert first_phi.lvalue in second_phi.rvalues
        assert len(second_phi.rvalues) == 2

        # return is for second phi
        assert len(return_value.values) == 1
        assert second_phi.lvalue in return_value.values


def test_issue_1748(slither_from_solidity_source):
    source = """
    contract Contract {
        uint[] arr;
        function foo(uint i) public {
            arr = [1];
        }
    }
    """
    with slither_from_solidity_source(source) as slither:
        c = slither.get_contract_from_name("Contract")[0]
        f = c.functions[0]
        operations = f.slithir_operations
        assign_op = operations[0]
        assert isinstance(assign_op, InitArray)


def test_issue_1776(slither_from_solidity_source):
    source = """
    contract Contract {
        function foo() public returns (uint) {
            uint[5][10][] memory arr = new uint[5][10][](2);
            return 0;
        }
    }
    """
    with slither_from_solidity_source(source) as slither:
        c = slither.get_contract_from_name("Contract")[0]
        f = c.functions[0]
        operations = f.slithir_operations
        new_op = operations[0]
        lvalue = new_op.lvalue
        lvalue_type = lvalue.type
        assert isinstance(lvalue_type, ArrayType)
        assert lvalue_type.is_dynamic
        lvalue_type1 = lvalue_type.type
        assert isinstance(lvalue_type1, ArrayType)
        assert not lvalue_type1.is_dynamic
        assert lvalue_type1.length_value.value == "10"
        lvalue_type2 = lvalue_type1.type
        assert isinstance(lvalue_type2, ArrayType)
        assert not lvalue_type2.is_dynamic
        assert lvalue_type2.length_value.value == "5"


def test_issue_1846_ternary_in_if(slither_from_solidity_source):
    source = """
    contract Contract {
        function foo(uint x) public returns (uint y) {
            if (x > 0) {
                y = x > 1 ? 2 : 3;
            } else {
                y = 4;
            }
        }
    }
    """
    with slither_from_solidity_source(source) as slither:
        c = slither.get_contract_from_name("Contract")[0]
        f = c.functions[0]
        node = f.nodes[1]
        assert node.type == NodeType.IF
        assert node.son_true.type == NodeType.IF
        assert node.son_false.type == NodeType.EXPRESSION


def test_issue_1846_ternary_in_ternary(slither_from_solidity_source):
    source = """
        contract Contract {
            function foo(uint x) public returns (uint y) {
                y = x > 0 ? x > 1 ? 2 : 3 : 4;
            }
        }
        """
    with slither_from_solidity_source(source) as slither:
        c = slither.get_contract_from_name("Contract")[0]
        f = c.functions[0]
        node = f.nodes[1]
        assert node.type == NodeType.IF
        assert node.son_true.type == NodeType.IF
        assert node.son_false.type == NodeType.EXPRESSION


def test_issue_2016(slither_from_solidity_source):
    source = """
    contract Contract {
        function test() external {
            int[] memory a = new int[](5);
        }
    }
    """
    with slither_from_solidity_source(source) as slither:
        c = slither.get_contract_from_name("Contract")[0]
        f = c.functions[0]
        operations = f.slithir_operations
        new_op = operations[0]
        assert isinstance(new_op, NewArray)
        lvalue = new_op.lvalue
        lvalue_type = lvalue.type
        assert isinstance(lvalue_type, ArrayType)
        assert lvalue_type.type == ElementaryType("int256")
        assert lvalue_type.is_dynamic


def test_issue_2210(slither_from_solidity_source):
    source = """
    contract C {
    function f (int x) public returns(int) {
        int h = 1;
        int k = 5;
        int[5] memory arr = [x, C.x, C.y, h - k, h + k];
    }
    int x= 4;
    int y = 5;
    }
    """
    with slither_from_solidity_source(source) as slither:
        c = slither.get_contract_from_name("C")[0]
        f = c.functions[0]
        operations = f.slithir_operations
        new_op = operations[6]
        assert isinstance(new_op, InitArray)
        lvalue = new_op.lvalue
        lvalue_type = lvalue.type
        assert isinstance(lvalue_type, ArrayType)
        assert lvalue_type.type == ElementaryType("int256")
        assert not lvalue_type.is_dynamic

    source2 = """
    contract X {
        function _toInts(uint256[] memory a) private pure returns (int256[] memory casted) {
            /// @solidity memory-safe-assembly
            assembly {
                casted := a
            }
        }
    }
    """
    with slither_from_solidity_source(source2) as slither:
        x = slither.get_contract_from_name("X")[0]
        f2 = x.functions[0]
        operations = f2.slithir_operations
        new_op2 = operations[0]
        assert isinstance(new_op2, Assignment)

        lvalue = new_op2.lvalue
        lvalue_type = lvalue.type
        assert isinstance(lvalue_type, ArrayType)
        assert lvalue_type.type == ElementaryType("int256")
        assert lvalue_type.is_dynamic

        rvalue_type = new_op2.rvalue.type
        assert isinstance(rvalue_type, ArrayType)
        assert rvalue_type.type == ElementaryType("uint256")
        assert rvalue_type.is_dynamic



import argparse
import logging
import sys
import os.path
import traceback

from .model import load_model
from .encode import parse_target, encode_contract

logging.basicConfig()
logger = logging.getLogger("Slither-simil")


def info(args: argparse.Namespace) -> None:
    try:
        model = args.model
        if os.path.isfile(model):
            model = load_model(model)
        else:
            model = None

        filename = args.filename
        contract, fname = parse_target(args.fname)

        if filename is None and contract is None and fname is None:
            logger.info("%s uses the following words:", args.model)
            for word in model.get_words():
                logger.info(word)
            sys.exit(0)

        if filename is None or contract is None or fname is None:
            logger.error("The encode mode requires filename, contract and fname parameters.")
            sys.exit(-1)

        irs = encode_contract(filename, **vars(args))
        if len(irs) == 0:
            sys.exit(-1)

        x = (filename, contract, fname)
        y = " ".join(irs[x])

        to_log = f"Function {fname} in contract {contract} is encoded as:"
        logger.info(to_log)
        logger.info(y)
        if model is not None:
            fvector = model.get_sentence_vector(y)
            logger.info(fvector)

    except Exception:
        to_log = f"Error in {args.filename}"
        logger.error(to_log)
        logger.error(traceback.format_exc())
        sys.exit(-1)
